# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Dict

from fastapi import FastAPI
from fastapi.testclient import TestClient
from pytest import fixture, mark, raises, skip
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.exceptions import GeneralError
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.querier import Querier
from supertokens_python.recipe import passwordless, session
from supertokens_python.recipe.passwordless.asyncio import (
    delete_email_for_user,
    delete_phone_number_for_user,
    get_user_by_email,
    get_user_by_id,
    get_user_by_phone_number,
    update_user,
)
from supertokens_python.recipe.passwordless.interfaces import (
    DeleteUserInfoOkResult,
    UpdateUserOkResult,
)
from supertokens_python.utils import is_version_gte

from tests.utils import clean_st, reset, setup_st, start_st


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@fixture(scope="function")
async def driver_config_client():
    app = FastAPI()
    app.add_middleware(get_middleware())

    return TestClient(app)


@mark.asyncio
async def test_passwordless_otp(driver_config_client: TestClient):
    user_input_code = None

    async def send_text_message(
        param: passwordless.CreateAndSendCustomTextMessageParameters, _: Dict[str, Any]
    ):
        nonlocal user_input_code
        user_input_code = param.user_input_code

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            passwordless.init(
                flow_type="USER_INPUT_CODE",
                contact_config=passwordless.ContactPhoneOnlyConfig(
                    create_and_send_custom_text_message=send_text_message
                ),
            ),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    version = await Querier.get_instance().get_api_version()
    if not is_version_gte(version, "2.11"):
        # If the version less than 2.11.0, passwordless OTP doesn't exist. So skip the test
        skip()

    create_code_json = driver_config_client.post(
        url="/auth/signinup/code",
        json={
            "phoneNumber": "+919494949494",
        },
    ).json()

    consume_code_json = driver_config_client.post(
        url="/auth/signinup/code/consume",
        json={
            "preAuthSessionId": create_code_json["preAuthSessionId"],
            "deviceId": create_code_json["deviceId"],
            "userInputCode": user_input_code,
        },
    ).json()

    consume_code_json["user"].pop("id")
    consume_code_json["user"].pop("time_joined")

    assert consume_code_json == {
        "status": "OK",
        "createdNewUser": True,
        "user": {"phoneNumber": "+919494949494"},
    }


@mark.asyncio
async def test_passworldless_delete_user_phone(driver_config_client: TestClient):
    text_code = None
    email_code = None

    async def send_text_message(
        param: passwordless.CreateAndSendCustomTextMessageParameters, _: Dict[str, Any]
    ):
        nonlocal text_code
        text_code = param.user_input_code

    async def send_email(
        param: passwordless.CreateAndSendCustomEmailParameters, _: Dict[str, Any]
    ):
        nonlocal email_code
        email_code = param.user_input_code

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            passwordless.init(
                flow_type="USER_INPUT_CODE",
                contact_config=passwordless.ContactEmailOrPhoneConfig(
                    create_and_send_custom_text_message=send_text_message,
                    create_and_send_custom_email=send_email,
                ),
            ),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    version = await Querier.get_instance().get_api_version()
    if not is_version_gte(version, "2.11"):
        # If the version less than 2.11, passwordless OTP doesn't exist. So skip the test
        skip()

    create_code_json = driver_config_client.post(
        url="/auth/signinup/code",
        json={
            "phoneNumber": "+919494949494",
        },
    ).json()

    consume_code_json = driver_config_client.post(
        url="/auth/signinup/code/consume",
        json={
            "preAuthSessionId": create_code_json["preAuthSessionId"],
            "deviceId": create_code_json["deviceId"],
            "userInputCode": text_code,
        },
    ).json()

    user_id = consume_code_json["user"]["id"]

    await update_user(user_id, "foo@example.com", "+919494949494")

    response = await delete_phone_number_for_user(user_id)
    assert isinstance(response, DeleteUserInfoOkResult)

    user = await get_user_by_phone_number("+919494949494")
    assert user is None

    user = await get_user_by_id(user_id)
    assert user is not None and user.phone_number is None


@mark.asyncio
async def test_passworldless_delete_user_email(driver_config_client: TestClient):
    text_code = None
    email_code = None

    async def send_text_message(
        param: passwordless.CreateAndSendCustomTextMessageParameters, _: Dict[str, Any]
    ):
        nonlocal text_code
        text_code = param.user_input_code

    async def send_email(
        param: passwordless.CreateAndSendCustomEmailParameters, _: Dict[str, Any]
    ):
        nonlocal email_code
        email_code = param.user_input_code

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            passwordless.init(
                flow_type="USER_INPUT_CODE",
                contact_config=passwordless.ContactEmailOrPhoneConfig(
                    create_and_send_custom_text_message=send_text_message,
                    create_and_send_custom_email=send_email,
                ),
            ),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    version = await Querier.get_instance().get_api_version()
    if not is_version_gte(version, "2.11"):
        # If the version less than 2.11, passwordless OTP doesn't exist. So skip the test
        skip()

    create_code_json = driver_config_client.post(
        url="/auth/signinup/code",
        json={
            "email": "hello@example.com",
        },
    ).json()

    consume_code_json = driver_config_client.post(
        url="/auth/signinup/code/consume",
        json={
            "preAuthSessionId": create_code_json["preAuthSessionId"],
            "deviceId": create_code_json["deviceId"],
            "userInputCode": email_code,
        },
    ).json()

    user_id = consume_code_json["user"]["id"]

    await update_user(user_id, "hello@example.com", "+919494949494")

    response = await delete_email_for_user(user_id)
    assert isinstance(response, DeleteUserInfoOkResult)

    user = await get_user_by_email("hello@example.com")
    assert user is None

    user = await get_user_by_id(user_id)
    assert user is not None and user.email is None


@mark.asyncio
async def test_passworldless_delete_user_email_and_phone_throws_error(
    driver_config_client: TestClient,
):
    text_code = None
    email_code = None

    async def send_text_message(
        param: passwordless.CreateAndSendCustomTextMessageParameters, _: Dict[str, Any]
    ):
        nonlocal text_code
        text_code = param.user_input_code

    async def send_email(
        param: passwordless.CreateAndSendCustomEmailParameters, _: Dict[str, Any]
    ):
        nonlocal email_code
        email_code = param.user_input_code

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            passwordless.init(
                flow_type="USER_INPUT_CODE",
                contact_config=passwordless.ContactEmailOrPhoneConfig(
                    create_and_send_custom_text_message=send_text_message,
                    create_and_send_custom_email=send_email,
                ),
            ),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    version = await Querier.get_instance().get_api_version()
    if not is_version_gte(version, "2.11"):
        # If the version less than 2.11, passwordless OTP doesn't exist. So skip the test
        skip()

    create_code_json = driver_config_client.post(
        url="/auth/signinup/code",
        json={
            "email": "hello@example.com",
        },
    ).json()

    consume_code_json = driver_config_client.post(
        url="/auth/signinup/code/consume",
        json={
            "preAuthSessionId": create_code_json["preAuthSessionId"],
            "deviceId": create_code_json["deviceId"],
            "userInputCode": email_code,
        },
    ).json()

    user_id = consume_code_json["user"]["id"]

    response = await update_user(user_id, "hello@example.com", "+919494949494")
    assert isinstance(response, UpdateUserOkResult)

    # Delete the email
    response = await delete_email_for_user(user_id)
    # Delete the phone number (Should raise exception because deleting both of them isn't allowed)
    with raises(GeneralError) as e:
        response = await delete_phone_number_for_user(user_id)

    assert e.value.args[0].endswith(
        "You cannot clear both email and phone number of a user\n"
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from os import environ
from typing import TYPE_CHECKING, Awaitable, Callable, List, Optional, Union

from supertokens_python.normalised_url_path import NormalisedURLPath
from supertokens_python.recipe_module import APIHandled, RecipeModule

from .api import (
    api_key_protector,
    handle_dashboard_api,
    handle_email_verify_token_post,
    handle_emailpassword_signin_api,
    handle_emailpassword_signout_api,
    handle_metadata_get,
    handle_metadata_put,
    handle_sessions_get,
    handle_user_delete,
    handle_user_email_verify_get,
    handle_user_email_verify_put,
    handle_user_get,
    handle_user_password_put,
    handle_user_put,
    handle_user_sessions_post,
    handle_users_count_get_api,
    handle_users_get_api,
    handle_validate_key_api,
)
from .api.implementation import APIImplementation
from .exceptions import SuperTokensDashboardError
from .interfaces import APIInterface, APIOptions
from .recipe_implementation import RecipeImplementation

if TYPE_CHECKING:
    from supertokens_python.framework.request import BaseRequest
    from supertokens_python.framework.response import BaseResponse
    from supertokens_python.supertokens import AppInfo
    from supertokens_python.types import APIResponse

from supertokens_python.exceptions import SuperTokensError, raise_general_exception

from .constants import (
    DASHBOARD_API,
    EMAIL_PASSSWORD_SIGNOUT,
    EMAIL_PASSWORD_SIGN_IN,
    USER_API,
    USER_EMAIL_VERIFY_API,
    USER_EMAIL_VERIFY_TOKEN_API,
    USER_METADATA_API,
    USER_PASSWORD_API,
    USER_SESSION_API,
    USERS_COUNT_API,
    USERS_LIST_GET_API,
    VALIDATE_KEY_API,
)
from .utils import (
    InputOverrideConfig,
    get_api_if_matched,
    is_api_path,
    validate_and_normalise_user_input,
)


class DashboardRecipe(RecipeModule):
    recipe_id = "dashboard"
    __instance = None

    def __init__(
        self,
        recipe_id: str,
        app_info: AppInfo,
        api_key: Union[str, None],
        override: Union[InputOverrideConfig, None] = None,
    ):
        super().__init__(recipe_id, app_info)
        self.config = validate_and_normalise_user_input(
            api_key,
            override,
        )
        recipe_implementation = RecipeImplementation()
        self.recipe_implementation = (
            recipe_implementation
            if self.config.override.functions is None
            else self.config.override.functions(recipe_implementation)
        )

        api_implementation = APIImplementation()
        self.api_implementation = (
            api_implementation
            if self.config.override.apis is None
            else self.config.override.apis(api_implementation)
        )

    def is_error_from_this_recipe_based_on_instance(self, err: Exception) -> bool:
        return isinstance(err, SuperTokensError) and (
            isinstance(err, SuperTokensDashboardError)
        )

    def get_apis_handled(self) -> List[APIHandled]:
        # Normally this array is used by the SDK to decide whether the recipe
        # handles a specific API path and method and then returns the ID.

        # However, for the dashboard recipe this logic is fully custom and handled inside the
        # `return_api_id_if_can_handle_request` method of this class. Since this array is never
        # used for this recipe, we simply return an empty array.

        return []

    async def handle_api_request(
        self,
        request_id: str,
        request: BaseRequest,
        path: NormalisedURLPath,
        method: str,
        response: BaseResponse,
    ) -> Optional[BaseResponse]:
        api_options = APIOptions(
            request,
            response,
            self.recipe_id,
            self.config,
            self.recipe_implementation,
            self.get_app_info(),
        )
        # For these APIs we dont need API key validation
        if request_id == DASHBOARD_API:
            return await handle_dashboard_api(self.api_implementation, api_options)
        if request_id == VALIDATE_KEY_API:
            return await handle_validate_key_api(self.api_implementation, api_options)
        if request_id == EMAIL_PASSWORD_SIGN_IN:
            return await handle_emailpassword_signin_api(
                self.api_implementation, api_options
            )

        # Do API key validation for the remaining APIs
        api_function: Optional[
            Callable[[APIInterface, APIOptions], Awaitable[APIResponse]]
        ] = None
        if request_id == USERS_LIST_GET_API:
            api_function = handle_users_get_api
        elif request_id == USERS_COUNT_API:
            api_function = handle_users_count_get_api
        elif request_id == USER_API:
            if method == "get":
                api_function = handle_user_get
            if method == "delete":
                api_function = handle_user_delete
            if method == "put":
                api_function = handle_user_put
        elif request_id == USER_EMAIL_VERIFY_API:
            if method == "get":
                api_function = handle_user_email_verify_get
            if method == "put":
                api_function = handle_user_email_verify_put
        elif request_id == USER_METADATA_API:
            if method == "get":
                api_function = handle_metadata_get
            if method == "put":
                api_function = handle_metadata_put
        elif request_id == USER_SESSION_API:
            if method == "get":
                api_function = handle_sessions_get
            if method == "post":
                api_function = handle_user_sessions_post
        elif request_id == USER_PASSWORD_API:
            api_function = handle_user_password_put
        elif request_id == USER_EMAIL_VERIFY_TOKEN_API:
            api_function = handle_email_verify_token_post
        elif request_id == EMAIL_PASSSWORD_SIGNOUT:
            api_function = handle_emailpassword_signout_api

        if api_function is not None:
            return await api_key_protector(
                self.api_implementation, api_options, api_function
            )

        return None

    async def handle_error(
        self, request: BaseRequest, err: SuperTokensError, response: BaseResponse
    ) -> BaseResponse:
        raise err

    def get_all_cors_headers(self) -> List[str]:
        return []

    @staticmethod
    def init(
        api_key: Union[str, None],
        override: Union[InputOverrideConfig, None] = None,
    ):
        def func(app_info: AppInfo):
            if DashboardRecipe.__instance is None:
                DashboardRecipe.__instance = DashboardRecipe(
                    DashboardRecipe.recipe_id,
                    app_info,
                    api_key,
                    override,
                )
                return DashboardRecipe.__instance
            raise Exception(
                None,
                "Dashboard recipe has already been initialised. Please check your code for bugs.",
            )

        return func

    @staticmethod
    def get_instance() -> DashboardRecipe:
        if DashboardRecipe.__instance is not None:
            return DashboardRecipe.__instance
        raise_general_exception(
            "Initialisation not done. Did you forget to call the SuperTokens.init function?"
        )

    @staticmethod
    def reset():
        if ("SUPERTOKENS_ENV" not in environ) or (
            environ["SUPERTOKENS_ENV"] != "testing"
        ):
            raise_general_exception("calling testing function in non testing env")
        DashboardRecipe.__instance = None

    def return_api_id_if_can_handle_request(
        self, path: NormalisedURLPath, method: str
    ) -> Union[str, None]:
        dashboard_bundle_path = self.app_info.api_base_path.append(
            NormalisedURLPath(DASHBOARD_API)
        )

        if is_api_path(path, self.app_info):
            return get_api_if_matched(path, method)

        if path.startswith(dashboard_bundle_path):
            return DASHBOARD_API

        return None



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from supertokens_python.framework import BaseResponse

if TYPE_CHECKING:
    from supertokens_python.recipe.dashboard.interfaces import (
        APIOptions,
        APIInterface,
    )

from supertokens_python.utils import default_user_context


async def handle_dashboard_api(
    api_implementation: APIInterface, api_options: APIOptions
) -> Optional[BaseResponse]:
    if api_implementation.dashboard_get is None:
        return None

    user_context = default_user_context(api_options.request)
    html_str = await api_implementation.dashboard_get(api_options, user_context)

    api_options.response.set_html_content(html_str)
    return api_options.response



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from supertokens_python.recipe.dashboard.interfaces import APIInterface, APIOptions

from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.normalised_url_path import NormalisedURLPath
from supertokens_python.querier import Querier
from supertokens_python.utils import send_200_response


async def handle_emailpassword_signin_api(_: APIInterface, api_options: APIOptions):
    body = await api_options.request.json()
    if body is None:
        raise_bad_input_exception("Please send body")
    email = body.get("email")
    password = body.get("password")

    if email is None or not isinstance(email, str):
        raise_bad_input_exception("Missing required parameter 'email'")
    if password is None or not isinstance(password, str):
        raise_bad_input_exception("Missing required parameter 'password'")
    response = await Querier.get_instance().send_post_request(
        NormalisedURLPath("/recipe/dashboard/signin"),
        {"email": email, "password": password},
    )

    if "status" in response and response["status"] == "OK":
        return send_200_response(
            {"status": "OK", "sessionId": response["sessionId"]}, api_options.response
        )
    if "status" in response and response["status"] == "INVALID_CREDENTIALS_ERROR":
        return send_200_response(
            {"status": "INVALID_CREDENTIALS_ERROR"},
            api_options.response,
        )
    if "status" in response and response["status"] == "USER_SUSPENDED_ERROR":
        return send_200_response(
            {"status": "USER_SUSPENDED_ERROR", "message": response["message"]},
            api_options.response,
        )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urlparse

from .utils import is_an_ip_address

if TYPE_CHECKING:
    pass
from .exceptions import raise_general_exception


class NormalisedURLDomain:
    def __init__(self, url: str):
        self.__value = normalise_domain_path_or_throw_error(url)

    def get_as_string_dangerous(self):
        return self.__value


def normalise_domain_path_or_throw_error(
    input_str: str, ignore_protocol: bool = False
) -> str:
    input_str = input_str.strip().lower()

    try:
        if (
            (not input_str.startswith("http://"))
            and (not input_str.startswith("https://"))
            and (not input_str.startswith("supertokens://"))
        ):
            raise Exception("converting to proper URL")
        url_obj = urlparse(input_str)

        if ignore_protocol:
            if url_obj.hostname is None:
                raise Exception("Should never come here")
            if url_obj.hostname.startswith("localhost") or is_an_ip_address(
                url_obj.hostname
            ):
                input_str = "http://" + url_obj.netloc
            else:
                input_str = "https://" + url_obj.netloc
        else:
            input_str = url_obj.scheme + "://" + url_obj.netloc

        return input_str
    except Exception:
        pass

    if input_str.startswith("/"):
        raise_general_exception("Please provide a valid domain name")

    if input_str.startswith("."):
        input_str = input_str[1:]

    if (
        ("." in input_str or input_str.startswith("localhost"))
        and (not input_str.startswith("http://"))
        and (not input_str.startswith("https://"))
    ):
        input_str = "https://" + input_str
        try:
            urlparse(input_str)
            return normalise_domain_path_or_throw_error(input_str, True)
        except Exception:
            pass
    raise_general_exception("Please provide a valid domain name")



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from base64 import b64decode
from json import dumps, loads
from textwrap import wrap
from typing import Any, Dict

from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature.pkcs1_15 import PKCS115_SigScheme
from supertokens_python.utils import utf_base64decode, utf_base64encode

_key_start = "-----BEGIN PUBLIC KEY-----\n"
_key_end = "\n-----END PUBLIC KEY-----"

"""
why separators is used in dumps:
- without it's use, output of dumps is: '{"alg": "RS256", "typ": "JWT", "version": "1"}'
- with it's use, output of dumps is: '{"alg":"RS256","typ":"JWT","version":"1"}'

we require the non-spaced version, else the base64 encoding string will end up different than required
"""
_allowed_headers = [
    utf_base64encode(
        dumps(
            {"alg": "RS256", "typ": "JWT", "version": "2"},
            separators=(",", ":"),
            sort_keys=True,
        )
    )
]


class ParsedJWTInfo:
    def __init__(
        self,
        raw_token_string: str,
        raw_payload: str,
        header: str,
        payload: Dict[str, Any],
        signature: str,
    ) -> None:
        self.raw_token_string = raw_token_string
        self.raw_payload = raw_payload
        self.header = header
        self.payload = payload
        self.signature = signature


def parse_jwt_without_signature_verification(jwt: str) -> ParsedJWTInfo:
    splitted_input = jwt.split(".")
    if len(splitted_input) != 3:
        raise Exception("invalid jwt")

    header, payload, signature = splitted_input
    if header not in _allowed_headers:
        raise Exception("jwt header mismatch")

    return ParsedJWTInfo(
        raw_token_string=jwt,
        raw_payload=payload,
        header=header,
        # Ideally we would only parse this after the signature verification is done
        # We do this at the start, since we want to check if a token can be a supertokens access token or not.
        payload=loads(utf_base64decode(payload)),
        signature=signature,
    )


def verify_jwt(info: ParsedJWTInfo, jwt_signing_public_key: str):
    public_key = RSA.import_key(
        _key_start + "\n".join(wrap(jwt_signing_public_key, width=64)) + _key_end
    )
    verifier = PKCS115_SigScheme(public_key)
    to_verify = SHA256.new((info.header + "." + info.raw_payload).encode("utf-8"))
    try:
        verifier.verify(to_verify, b64decode(info.signature.encode("utf-8")))
    except BaseException:
        raise Exception("jwt verification failed")



from typing import List, Any, Dict, Union, Optional
from unittest.mock import patch

from fastapi import FastAPI, Depends
from pytest import fixture, mark
from starlette.requests import Request
from starlette.testclient import TestClient

from supertokens_python import init
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.recipe import session
from supertokens_python.recipe.session.asyncio import create_new_session
from supertokens_python.recipe.session.exceptions import (
    raise_invalid_claims_exception,
    ClaimValidationError,
)
from supertokens_python.recipe.session.framework.fastapi import verify_session
from supertokens_python.recipe.session.interfaces import (
    RecipeInterface,
    SessionClaimValidator,
    JSONObject,
    ClaimValidationResult,
    SessionContainer,
    ClaimsValidationResult,
)
from supertokens_python.recipe.session.session_class import Session
from tests.sessions.claims.utils import TrueClaim, NoneClaim
from tests.utils import (
    setup_function,
    teardown_function,
    start_st,
    AsyncMock,
    MagicMock,
    st_init_common_args,
)

_ = setup_function  # type:ignore
_ = teardown_function  # type:ignore
_ = start_st  # type:ignore

pytestmark = mark.asyncio


def st_init_generator_with_overriden_global_validators(
    validators: List[SessionClaimValidator],
):
    def session_function_override(oi: RecipeInterface) -> RecipeInterface:
        async def new_get_global_claim_validators(
            _user_id: str,
            _claim_validators_added_by_other_recipes: List[SessionClaimValidator],
            _user_context: Dict[str, Any],
        ):
            return validators

        oi.get_global_claim_validators = new_get_global_claim_validators
        return oi

    return {
        **st_init_common_args,
        "recipe_list": [
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
                override=session.InputOverrideConfig(
                    functions=session_function_override
                ),
            )
        ],
    }


def st_init_generator_with_claim_validator(claim_validator: SessionClaimValidator):
    def session_function_override(oi: RecipeInterface) -> RecipeInterface:
        async def new_get_global_claim_validators(
            _user_id: str,
            claim_validators_added_by_other_recipes: List[SessionClaimValidator],
            _user_context: Dict[str, Any],
        ):
            return [*claim_validators_added_by_other_recipes, claim_validator]

        oi.get_global_claim_validators = new_get_global_claim_validators
        return oi

    return {
        **st_init_common_args,
        "recipe_list": [
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
                override=session.InputOverrideConfig(
                    functions=session_function_override
                ),
            )
        ],
    }


class AlwaysValidValidator(SessionClaimValidator):
    def __init__(self):
        super().__init__("always-valid-validator")
        self.claim = TrueClaim

    async def validate(
        self, payload: JSONObject, user_context: Union[Dict[str, Any], None] = None
    ) -> ClaimValidationResult:
        return ClaimValidationResult(True)

    def should_refetch(self, payload: JSONObject, user_context: Dict[str, Any]):
        return True


class AlwaysInvalidValidator(SessionClaimValidator):
    def __init__(self, reason: Optional[Dict[str, Any]]):
        super().__init__("always-invalid-validator")
        self.claim = TrueClaim
        self.reason = reason

    async def validate(
        self, payload: JSONObject, user_context: Union[Dict[str, Any], None] = None
    ) -> ClaimValidationResult:
        return ClaimValidationResult(is_valid=False, reason=self.reason)

    def should_refetch(self, payload: JSONObject, user_context: Dict[str, Any]) -> bool:
        return True


@fixture(scope="function")
async def fastapi_client():
    app = FastAPI()
    app.add_middleware(get_middleware())

    @app.post("/login")
    async def _login(request: Request):  # type: ignore
        user_id = "userId"
        await create_new_session(request, user_id, {}, {})
        return {"userId": user_id}

    @app.post("/create-with-claim")
    async def _create_with_claim(request: Request):  # type: ignore
        user_id = "userId"
        _ = await create_new_session(request, user_id, {}, {})
        key: str = (await request.json())["key"]
        # PrimitiveClaim(key, fetch_value="Value").add_to_session(session, "value")
        return {"userId": key}

    @app.get("/default-claims")
    async def default_claims(s: SessionContainer = Depends(verify_session())):  # type: ignore
        return {"handle": s.get_handle()}

    no_claims_verify_session = verify_session(
        override_global_claim_validators=lambda _, __, ___: []  # type: ignore
    )

    @app.get("/no-claims")
    async def no_claims(s: SessionContainer = Depends(no_claims_verify_session)):  # type: ignore
        return {"handle": s.get_handle()}

    refetched_claims_verify_session = verify_session(
        override_global_claim_validators=lambda _, __, ___: [  # type: ignore
            TrueClaim.validators.has_value(True)
        ]
    )

    @app.get("/refetched-claim")
    async def refetched_claim(  # type: ignore
        s: SessionContainer = Depends(refetched_claims_verify_session),
    ):
        return {"handle": s.get_handle()}

    refetched_claims_false_verify_session = verify_session(
        override_global_claim_validators=lambda _, __, ___: [  # type: ignore
            TrueClaim.validators.has_value(False)
        ]
    )

    @app.get("/refetched-claim-false")
    async def refetched_claim2(  # type: ignore
        s: SessionContainer = Depends(refetched_claims_false_verify_session),
    ):
        return {"handle": s.get_handle()}

    class CustomValidator(SessionClaimValidator):
        def __init__(self, is_valid: bool):
            super().__init__("test_id")
            self.claim = TrueClaim
            self.is_valid = is_valid

        async def validate(
            self, payload: JSONObject, user_context: Union[Dict[str, Any], None] = None
        ) -> ClaimValidationResult:
            if self.is_valid:
                return ClaimValidationResult(is_valid=True)
            return ClaimValidationResult(
                is_valid=False, reason={"message": "test_reason"}
            )

        def should_refetch(self, payload: JSONObject, user_context: Dict[str, Any]):
            return True

    refetched_claims_verify_session_is_valid_false = verify_session(
        override_global_claim_validators=lambda _, __, ___: [  # type: ignore
            CustomValidator(is_valid=False)
        ]
    )

    @app.get("/refetched-claim-isvalid-false")
    async def refetched_claim3(  # type: ignore
        s: SessionContainer = Depends(refetched_claims_verify_session_is_valid_false),
    ):
        return {"handle": s.get_handle()}

    refetched_claims_verify_session_is_valid_true = verify_session(
        override_global_claim_validators=lambda _, __, ___: [  # type: ignore
            CustomValidator(is_valid=True)
        ]
    )

    @app.get("/refetched-claim-isvalid-true")
    async def refetched_claim4(  # type: ignore
        s: SessionContainer = Depends(refetched_claims_verify_session_is_valid_true),
    ):
        return {"handle": s.get_handle()}

    return TestClient(app)


def create_session(fastapi_client: TestClient):
    res = fastapi_client.post("create-with-claim", json={"key": "something"})
    return res


async def test_should_allow_without_claims_required_or_present(
    fastapi_client: TestClient,
):
    st_init_args = {
        **st_init_common_args,
        "recipe_list": [
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            )
        ],
    }
    init(**st_init_args)  # type: ignore
    start_st()

    create_session(fastapi_client)

    res = fastapi_client.get("/default-claims")
    assert res.status_code == 200
    assert "-" in res.json()["handle"]


async def test_should_allow_with_claim_valid_after_refetching(
    fastapi_client: TestClient,
):
    st_init_args = st_init_generator_with_claim_validator(
        TrueClaim.validators.has_value(True)
    )
    init(**st_init_args)  # type: ignore
    start_st()

    create_session(fastapi_client)

    response = fastapi_client.get("/default-claims")
    assert response.status_code == 200


async def test_should_reject_with_claim_required_but_not_added(
    fastapi_client: TestClient,
):
    st_init_args = st_init_generator_with_claim_validator(
        NoneClaim.validators.has_value(True)  # type: ignore
    )
    init(**st_init_args)  # type: ignore
    start_st()

    create_session(fastapi_client)

    response = fastapi_client.get("/default-claims")
    assert response.status_code == 403
    assert response.json() == {
        "message": "invalid claim",
        "claimValidationErrors": [
            {
                "id": "st-none",
                "reason": {
                    "message": "value does not exist",
                    "expectedValue": True,
                    "actualValue": None,
                },
            }
        ],
    }


async def test_should_allow_with_custom_validator_returning_true(
    fastapi_client: TestClient,
):
    custom_validator = AlwaysValidValidator()

    st_init_args = st_init_generator_with_claim_validator(custom_validator)
    init(**st_init_args)  # type: ignore
    start_st()

    create_session(fastapi_client)
    res = fastapi_client.get("/default-claims")
    assert res.status_code == 200
    assert "-" in res.json()["handle"]


async def test_should_reject_with_custom_validator_returning_false_without_reason(
    fastapi_client: TestClient,
):
    custom_validator = AlwaysInvalidValidator(reason=None)

    st_init_args = st_init_generator_with_claim_validator(custom_validator)
    init(**st_init_args)  # type: ignore
    start_st()

    create_session(fastapi_client)
    response = fastapi_client.get("/default-claims")
    assert response.status_code == 403
    assert response.json() == {
        "message": "invalid claim",
        "claimValidationErrors": [{"id": "always-invalid-validator"}],
    }


async def test_should_reject_with_custom_validator_returning_false_with_reason(
    fastapi_client: TestClient,
):
    custom_validator = AlwaysInvalidValidator(reason={"message": "foo"})

    st_init_args = st_init_generator_with_claim_validator(custom_validator)
    init(**st_init_args)  # type: ignore
    start_st()

    create_session(fastapi_client)
    response = fastapi_client.get("/default-claims")
    assert response.status_code == 403
    assert response.json() == {
        "message": "invalid claim",
        "claimValidationErrors": [
            {"id": "always-invalid-validator", "reason": {"message": "foo"}}
        ],
    }


async def test_should_reject_if_assert_claims_returns_an_error(
    fastapi_client: TestClient,
):
    validator = AlwaysValidValidator()
    st_init_args = st_init_generator_with_overriden_global_validators([validator])
    init(**st_init_args)  # type: ignore
    start_st()

    recipe_implementation_mock = AsyncMock()
    session_config_mock = MagicMock()
    s = Session(
        recipe_implementation_mock,
        session_config_mock,
        "test_access_token",
        "test_session_handle",
        "test_user_id",
        {},
        "cookie",
    )

    with patch.object(Session, "assert_claims", wraps=s.assert_claims) as mock:
        mock.side_effect = lambda _, __: raise_invalid_claims_exception(  # type: ignore
            "INVALID_CLAIM",
            [
                ClaimValidationError(
                    "test_id",
                    {"message": "test_reason"},
                )
            ],
        )
        create_session(fastapi_client)
        response = fastapi_client.get("/default-claims")
        assert response.status_code == 403
        assert response.json() == {
            "message": "invalid claim",
            "claimValidationErrors": [
                {"id": "test_id", "reason": {"message": "test_reason"}}
            ],
        }


async def test_should_allow_if_assert_claims_returns_no_error(
    fastapi_client: TestClient,
):
    validator = AlwaysValidValidator()
    st_init_args = st_init_generator_with_overriden_global_validators([validator])
    init(**st_init_args)  # type: ignore
    start_st()

    recipe_impl_mock = AsyncMock()
    recipe_impl_mock.validate_claims.return_value = ClaimsValidationResult(  # type: ignore
        invalid_claims=[]
    )
    session_config_mock = MagicMock()

    s = Session(
        recipe_impl_mock,
        session_config_mock,
        "test_access_token",
        "test_session_handle",
        "test_user_id",
        {},
        "cookie",
    )

    with patch.object(Session, "assert_claims", wraps=s.assert_claims) as mock:
        create_session(fastapi_client)
        response = fastapi_client.get("/default-claims")
        assert response.status_code == 200
        assert "-" in response.json()["handle"]
        mock.assert_called_once()
        (validators, ctx), _ = mock.call_args_list[0]
        assert validators == [validator]
        assert ctx["_default"]["request"]
        recipe_impl_mock.validate_claims.assert_called_once_with(  # type: ignore
            "test_user_id", {}, [validator], ctx
        )


# With override_global_claim_validators:


async def test_should_allow_with_empty_list_as_override(fastapi_client: TestClient):
    # Despite providing a SessionClaimValidator, the validator should not be called
    # because in the endpoint "no-claims" we remove all the validators
    # by passing override_global_claim_validators to verify_session()
    st_init_args = st_init_generator_with_claim_validator(
        NoneClaim.validators.has_value(True)
    )
    init(**st_init_args)  # type: ignore
    start_st()

    create_session(fastapi_client)
    res = fastapi_client.get("/no-claims")
    assert res.status_code == 200


async def test_should_allow_with_refetched_claim(fastapi_client: TestClient):
    # This gets overriden by TrueClaim.validators.has_value(True)
    # in refetched-claim-false via override_global_claim_validators param
    # passed to verify_session()
    st_init_args = st_init_generator_with_claim_validator(
        NoneClaim.validators.has_value(True)
    )
    init(**st_init_args)  # type: ignore
    start_st()

    create_session(fastapi_client)
    res = fastapi_client.get("/refetched-claim")
    assert res.status_code == 200


async def test_should_reject_with_invalid_refetched_claim(fastapi_client: TestClient):
    # This gets overriden by TrueClaim.validators.has_value(False)
    # in refetched-claim-false via override_global_claim_validators param
    # passed to verify_session()
    st_init_args = st_init_generator_with_claim_validator(
        NoneClaim.validators.has_value(True)
    )
    init(**st_init_args)  # type: ignore
    start_st()

    create_session(fastapi_client)
    res = fastapi_client.get("/refetched-claim-false")
    assert res.status_code == 403
    assert res.json() == {
        "message": "invalid claim",
        "claimValidationErrors": [
            {
                "id": "st-true",
                "reason": {
                    "message": "wrong value",
                    "expectedValue": False,
                    "actualValue": True,
                },
            }
        ],
    }


async def test_should_reject_with_custom_claim_returning_false(
    fastapi_client: TestClient,
):
    # This gets overriden by override_global_claim_validators passed to verify_session()
    # in "/refetched-claim-isvalid-false" api
    cv = AlwaysInvalidValidator(
        reason={"message": "does not matter because of override"}
    )
    st_init_args = st_init_generator_with_claim_validator(cv)
    init(**st_init_args)  # type: ignore
    start_st()

    create_session(fastapi_client)
    res = fastapi_client.get("/refetched-claim-isvalid-false")
    assert res.status_code == 403
    assert res.json() == {
        "message": "invalid claim",
        "claimValidationErrors": [
            {"id": "test_id", "reason": {"message": "test_reason"}}
        ],
    }


async def test_should_allow_with_custom_claim_returning_true(
    fastapi_client: TestClient,
):
    # This gets overriden by override_global_claim_validators passed to verify_session()
    # in "/refetched-claim-isvalid-true" api
    cv = AlwaysValidValidator()
    st_init_args = st_init_generator_with_claim_validator(cv)
    init(**st_init_args)  # type: ignore
    start_st()

    create_session(fastapi_client)
    res = fastapi_client.get("/refetched-claim-isvalid-true")
    assert res.status_code == 200
    assert "-" in res.json()["handle"]



# import asyncio
# import time
# from os import getenv
# from subprocess import PIPE, STDOUT, Popen, TimeoutExpired
# from unittest import TestCase

# import nest_asyncio  # type: ignore
# from supertokens_python.ingredients.emaildelivery.services.smtp import (
#     GetContentResult, SMTPServiceConfig, SMTPServiceConfigFrom, Transporter)


# class TransporterTests(TestCase):
#     def test_transporter(self):  # pylint: disable=no-self-use
#         local_insecure_smtpd_config_without_auth = SMTPServiceConfig(
#             host="localhost",
#             from_=SMTPServiceConfigFrom("Foo bar", "foo@example.com"),
#             port=1025,
#             secure=False,
#         )

#         transporter = Transporter(
#             smtp_settings=local_insecure_smtpd_config_without_auth,
#         )

#         content = GetContentResult(
#             body="<h1>Hello world</h1>",
#             subject='Greetings',
#             to_email='bar@example.com',
#             is_html=True,
#         )

#         command = "python3 -u -m smtpd -c DebuggingServer -n localhost:1025"
#         proc = Popen(command.split(), stdout=PIPE, stderr=STDOUT)  # Starts a SMTP daemon

#         is_sent = False

#         def send_email():
#             nonlocal is_sent
#             loop = asyncio.get_event_loop()
#             nest_asyncio.apply(loop)  # type: ignore
#             loop.run_until_complete(transporter.send_email(content, {}))
#             is_sent = True

#         try:
#             time.sleep(0.2)
#             send_email()
#         finally:
#             proc.terminate()
#             try:
#                 out, _ = proc.communicate(timeout=0.5)
#                 out = out.decode('utf-8').replace("\\n", "\n")
#                 assert out != ""
#                 assert out == """---------- MESSAGE FOLLOWS ----------
# b'Content-Type: text/html; charset="us-ascii"'
# b'MIME-Version: 1.0'
# b'Content-Transfer-Encoding: 7bit'
# b'From: Foo bar <foo@example.com>'
# b'To: bar@example.com'
# b'Subject: Greetings'
# b'X-Peer: 127.0.0.1'
# b''
# b'<h1>Hello world</h1>'
# ------------ END MESSAGE ------------
# """
#             except TimeoutExpired:
#                 raise Exception('subprocess did not terminate in time')

#     def test_transporter_with_gmail(self):  # pylint: disable=no-self-use
#         email = getenv("TEST_GMAIL_EMAIL")
#         password = getenv("TEST_GMAIL_PASS")

#         if not (email and password):
#             # Skip test if env vars aren't provided
#             return

#         real_gmail_smtp_config = SMTPServiceConfig(
#             host="smtp.gmail.com",
#             from_=SMTPServiceConfigFrom("ST Demo", email),
#             password=password,
#             port=465,  # alternatively, port=587, secure=False should also work
#             secure=True,
#         )

#         transporter = Transporter(
#             smtp_settings=real_gmail_smtp_config,
#         )

#         content = GetContentResult(
#             body="<h1>Hello world</h1>",
#             subject='Greetings',
#             to_email=email,
#             is_html=True,
#         )

#         loop = asyncio.get_event_loop()
#         nest_asyncio.apply(loop)  # type: ignore
#         loop.run_until_complete(transporter.send_email(content, {}))



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Union

import supertokens_python.recipe.emailpassword.interfaces as EPInterfaces

if TYPE_CHECKING:
    from supertokens_python.querier import Querier

from supertokens_python.recipe.emailpassword.recipe_implementation import (
    RecipeImplementation as EmailPasswordImplementation,
)
from supertokens_python.recipe.thirdparty.recipe_implementation import (
    RecipeImplementation as ThirdPartyImplementation,
)

from ..interfaces import (
    CreateResetPasswordOkResult,
    CreateResetPasswordWrongUserIdError,
    EmailPasswordSignInOkResult,
    EmailPasswordSignInWrongCredentialsError,
    EmailPasswordSignUpEmailAlreadyExistsError,
    EmailPasswordSignUpOkResult,
    RecipeInterface,
    ResetPasswordUsingTokenInvalidTokenError,
    ResetPasswordUsingTokenOkResult,
    ThirdPartySignInUpOkResult,
    UpdateEmailOrPasswordEmailAlreadyExistsError,
    UpdateEmailOrPasswordOkResult,
    UpdateEmailOrPasswordUnknownUserIdError,
)
from ..types import User
from .email_password_recipe_implementation import (
    RecipeImplementation as DerivedEmailPasswordImplementation,
)
from .third_party_recipe_implementation import (
    RecipeImplementation as DerivedThirdPartyImplementation,
)


class RecipeImplementation(RecipeInterface):
    def __init__(
        self, emailpassword_querier: Querier, thirdparty_querier: Union[Querier, None]
    ):
        super().__init__()
        emailpassword_implementation = EmailPasswordImplementation(
            emailpassword_querier
        )

        self.ep_get_user_by_id = emailpassword_implementation.get_user_by_id
        self.ep_get_user_by_email = emailpassword_implementation.get_user_by_email
        self.ep_create_reset_password_token = (
            emailpassword_implementation.create_reset_password_token
        )
        self.ep_reset_password_using_token = (
            emailpassword_implementation.reset_password_using_token
        )
        self.ep_sign_in = emailpassword_implementation.sign_in
        self.ep_sign_up = emailpassword_implementation.sign_up
        self.ep_update_email_or_password = (
            emailpassword_implementation.update_email_or_password
        )

        derived_ep = DerivedEmailPasswordImplementation(self)
        emailpassword_implementation.create_reset_password_token = (
            derived_ep.create_reset_password_token
        )
        emailpassword_implementation.get_user_by_email = derived_ep.get_user_by_email
        emailpassword_implementation.get_user_by_id = derived_ep.get_user_by_id
        emailpassword_implementation.reset_password_using_token = (
            derived_ep.reset_password_using_token
        )
        emailpassword_implementation.sign_in = derived_ep.sign_in
        emailpassword_implementation.sign_up = derived_ep.sign_up
        emailpassword_implementation.update_email_or_password = (
            derived_ep.update_email_or_password
        )

        self.tp_get_user_by_id = None
        self.tp_get_users_by_email = None
        self.tp_get_user_by_thirdparty_info = None
        self.tp_sign_in_up = None
        if thirdparty_querier is not None:
            thirdparty_implementation = ThirdPartyImplementation(thirdparty_querier)
            self.tp_get_user_by_id = thirdparty_implementation.get_user_by_id
            self.tp_get_users_by_email = thirdparty_implementation.get_users_by_email
            self.tp_get_user_by_thirdparty_info = (
                thirdparty_implementation.get_user_by_thirdparty_info
            )
            self.tp_sign_in_up = thirdparty_implementation.sign_in_up

            derived_tp = DerivedThirdPartyImplementation(self)
            thirdparty_implementation.get_user_by_id = derived_tp.get_user_by_id
            thirdparty_implementation.get_users_by_email = derived_tp.get_users_by_email
            thirdparty_implementation.get_user_by_thirdparty_info = (
                derived_tp.get_user_by_thirdparty_info
            )
            thirdparty_implementation.sign_in_up = derived_tp.sign_in_up

    async def get_user_by_id(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[User, None]:
        ep_user = await self.ep_get_user_by_id(user_id, user_context)

        if ep_user is not None:
            return User(
                user_id=ep_user.user_id,
                email=ep_user.email,
                time_joined=ep_user.time_joined,
                third_party_info=None,
            )

        if self.tp_get_user_by_id is None:
            return None

        tp_user = await self.tp_get_user_by_id(user_id, user_context)
        if tp_user is None:
            return None
        return User(
            user_id=tp_user.user_id,
            email=tp_user.email,
            time_joined=tp_user.time_joined,
            third_party_info=tp_user.third_party_info,
        )

    async def get_users_by_email(
        self, email: str, user_context: Dict[str, Any]
    ) -> List[User]:
        result: List[User] = []
        ep_user = await self.ep_get_user_by_email(email, user_context)

        if ep_user is not None:
            result.append(
                User(
                    user_id=ep_user.user_id,
                    email=ep_user.email,
                    time_joined=ep_user.time_joined,
                    third_party_info=None,
                )
            )

        if self.tp_get_users_by_email is None:
            return result

        tp_users = await self.tp_get_users_by_email(email, user_context)

        for tp_user in tp_users:
            result.append(
                User(
                    user_id=tp_user.user_id,
                    email=tp_user.email,
                    time_joined=tp_user.time_joined,
                    third_party_info=tp_user.third_party_info,
                )
            )

        return result

    async def get_user_by_thirdparty_info(
        self,
        third_party_id: str,
        third_party_user_id: str,
        user_context: Dict[str, Any],
    ) -> Union[User, None]:
        if self.tp_get_user_by_thirdparty_info is None:
            return None
        tp_user = await self.tp_get_user_by_thirdparty_info(
            third_party_id, third_party_user_id, user_context
        )

        if tp_user is None:
            return None

        return User(
            user_id=tp_user.user_id,
            email=tp_user.email,
            time_joined=tp_user.time_joined,
            third_party_info=tp_user.third_party_info,
        )

    async def thirdparty_sign_in_up(
        self,
        third_party_id: str,
        third_party_user_id: str,
        email: str,
        user_context: Dict[str, Any],
    ) -> ThirdPartySignInUpOkResult:
        if self.tp_sign_in_up is None:
            raise Exception("No thirdparty provider configured")
        result = await self.tp_sign_in_up(
            third_party_id, third_party_user_id, email, user_context
        )
        return ThirdPartySignInUpOkResult(
            User(
                result.user.user_id,
                result.user.email,
                result.user.time_joined,
                result.user.third_party_info,
            ),
            result.created_new_user,
        )

    async def emailpassword_sign_in(
        self, email: str, password: str, user_context: Dict[str, Any]
    ) -> Union[EmailPasswordSignInOkResult, EmailPasswordSignInWrongCredentialsError]:
        result = await self.ep_sign_in(email, password, user_context)
        if isinstance(result, EPInterfaces.SignInOkResult):
            return EmailPasswordSignInOkResult(
                User(
                    result.user.user_id,
                    result.user.email,
                    result.user.time_joined,
                    None,
                )
            )
        return result

    async def emailpassword_sign_up(
        self, email: str, password: str, user_context: Dict[str, Any]
    ) -> Union[EmailPasswordSignUpOkResult, EmailPasswordSignUpEmailAlreadyExistsError]:
        result = await self.ep_sign_up(email, password, user_context)
        if isinstance(result, EPInterfaces.SignUpOkResult):
            return EmailPasswordSignUpOkResult(
                User(
                    result.user.user_id,
                    result.user.email,
                    result.user.time_joined,
                    None,
                )
            )
        return result

    async def create_reset_password_token(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[CreateResetPasswordOkResult, CreateResetPasswordWrongUserIdError]:
        return await self.ep_create_reset_password_token(user_id, user_context)

    async def reset_password_using_token(
        self, token: str, new_password: str, user_context: Dict[str, Any]
    ) -> Union[
        ResetPasswordUsingTokenOkResult, ResetPasswordUsingTokenInvalidTokenError
    ]:
        return await self.ep_reset_password_using_token(
            token, new_password, user_context
        )

    async def update_email_or_password(
        self,
        user_id: str,
        email: Union[None, str],
        password: Union[None, str],
        user_context: Dict[str, Any],
    ) -> Union[
        UpdateEmailOrPasswordOkResult,
        UpdateEmailOrPasswordEmailAlreadyExistsError,
        UpdateEmailOrPasswordUnknownUserIdError,
    ]:
        user = await self.get_user_by_id(user_id, user_context)
        if user is None:
            return UpdateEmailOrPasswordUnknownUserIdError()
        if user.third_party_info is not None:
            raise Exception(
                "Cannot update email or password of a user who signed up using third party login."
            )
        return await self.ep_update_email_or_password(
            user_id, email, password, user_context
        )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import cast

import pytest
from typing_extensions import Literal

from pytest import mark, skip
from supertokens_python import init
from supertokens_python.interfaces import (
    DeleteUserIdMappingOkResult,
    CreateUserIdMappingOkResult,
    UnknownMappingError,
)
from supertokens_python.querier import Querier
from supertokens_python.recipe.emailpassword.interfaces import SignUpOkResult
from supertokens_python.recipe.usermetadata.asyncio import update_user_metadata
from supertokens_python.utils import is_version_gte
from tests.utils import clean_st, reset, setup_st, start_st
from supertokens_python.recipe.emailpassword.asyncio import sign_up
from .utils import st_config
from supertokens_python.asyncio import (
    create_user_id_mapping,
    get_user_id_mapping,
    delete_user_id_mapping,
)


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


pytestmark = mark.asyncio

USER_TYPE = Literal["SUPERTOKENS", "EXTERNAL", "ANY"]


@mark.parametrize("user_type", ["SUPERTOKENS", "EXTERNAL", "ANY"])
async def test_delete_user_id_mapping(user_type: USER_TYPE):
    init(**st_config)  # type: ignore
    start_st()

    version = await Querier.get_instance().get_api_version()
    if not is_version_gte(version, "2.15"):
        skip()

    # Create ST User
    sign_up_res = await sign_up("test@example.com", "password")
    assert isinstance(sign_up_res, SignUpOkResult)

    supertokens_user_id = sign_up_res.user.user_id
    external_user_id = "externalId"
    external_id_info = "externalIdInfo"

    # Create user id mapping
    res = await create_user_id_mapping(
        supertokens_user_id, external_user_id, external_id_info
    )
    assert isinstance(res, CreateUserIdMappingOkResult)

    user_id_to_delete = {
        "SUPERTOKENS": supertokens_user_id,
        "EXTERNAL": external_user_id,
        "ANY": external_user_id,
    }[user_type]

    # Delete the mapping
    res = await delete_user_id_mapping(user_id_to_delete, user_type)
    assert isinstance(res, DeleteUserIdMappingOkResult)
    assert res.did_mapping_exist is True

    # Ensure that the mapping is deleted
    res = await get_user_id_mapping(user_id_to_delete, user_type)
    assert isinstance(res, UnknownMappingError)


async def test_delete_user_id_mapping_without_and_with_force():
    init(**st_config)  # type: ignore
    start_st()

    version = await Querier.get_instance().get_api_version()
    if not is_version_gte(version, "2.15"):
        skip()

    # Create a user:
    sign_up_res = await sign_up("test@example.com", "testPass123")
    assert isinstance(sign_up_res, SignUpOkResult)

    supertokens_user_id = sign_up_res.user.user_id
    external_user_id = "externalId"
    external_user_info = "externalIdInfo"

    res = await create_user_id_mapping(
        supertokens_user_id, external_user_id, external_user_info
    )
    assert isinstance(res, CreateUserIdMappingOkResult)

    # Add metadata to the user:
    test_metadata = {"role": "admin"}
    await update_user_metadata(external_user_id, test_metadata)

    # Without force:
    with pytest.raises(Exception) as e:
        await delete_user_id_mapping(external_user_id, "EXTERNAL")
    assert str(e.value).endswith("UserId is already in use in UserMetadata recipe\n")

    # With force = False:
    with pytest.raises(Exception) as e:
        await delete_user_id_mapping(external_user_id, "EXTERNAL", force=False)
    assert str(e.value).endswith("UserId is already in use in UserMetadata recipe\n")

    # With force = True:
    res = await delete_user_id_mapping(external_user_id, "EXTERNAL", force=True)
    assert isinstance(res, DeleteUserIdMappingOkResult)
    assert res.did_mapping_exist is True

    # Check that User ID Mapping doesn't exist:
    get_user_id_mapping_res = await get_user_id_mapping(
        supertokens_user_id, "SUPERTOKENS"
    )
    assert isinstance(get_user_id_mapping_res, UnknownMappingError)


async def test_delete_unknown_user_id_mapping():
    init(**st_config)  # type: ignore
    start_st()

    version = await Querier.get_instance().get_api_version()
    if not is_version_gte(version, "2.15"):
        skip()

    for t in ["SUPERTOKENS", "EXTERNAL", "ANY"]:
        user_type = cast(USER_TYPE, t)
        res = await delete_user_id_mapping("non_existing_user_id", user_type)
        assert isinstance(res, DeleteUserIdMappingOkResult)
        assert res.did_mapping_exist is False



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, List, Optional, Union
from urllib.parse import urlparse

from typing_extensions import Literal

from supertokens_python.exceptions import raise_general_exception
from supertokens_python.framework import BaseResponse
from supertokens_python.normalised_url_path import NormalisedURLPath
from supertokens_python.recipe.openid import (
    InputOverrideConfig as OpenIdInputOverrideConfig,
)
from supertokens_python.utils import (
    is_an_ip_address,
    resolve,
    send_non_200_response,
    send_non_200_response_with_message,
)

from ...types import MaybeAwaitable
from .constants import AUTH_MODE_HEADER_KEY, SESSION_REFRESH
from .cookie_and_header import clear_session_from_all_token_transfer_methods
from .exceptions import ClaimValidationError
from .with_jwt.constants import (
    ACCESS_TOKEN_PAYLOAD_JWT_PROPERTY_NAME_KEY,
    JWT_RESERVED_KEY_USE_ERROR_MESSAGE,
)

if TYPE_CHECKING:
    from supertokens_python.framework import BaseRequest
    from supertokens_python.supertokens import AppInfo

    from .interfaces import (
        APIInterface,
        RecipeInterface,
        SessionContainer,
        SessionClaimValidator,
    )
    from .recipe import SessionRecipe

from supertokens_python.logger import log_debug_message

HUNDRED_YEARS_IN_MS = 3153600000000


def normalise_session_scope(session_scope: str) -> str:
    def helper(scope: str) -> str:
        scope = scope.strip()

        if scope.startswith("."):
            scope = scope[1:]

        if (not scope.startswith("https://")) and (not scope.startswith("http://")):
            scope = "http://" + scope

        try:
            url_obj = urlparse(scope)
            if url_obj.hostname is None:
                raise Exception("Should not come here")
            scope = url_obj.hostname

            if scope.startswith("."):
                scope = scope[1:]

            return scope
        except Exception:
            raise_general_exception("Please provide a valid sessionScope")

    no_dot_normalised = helper(session_scope)
    if no_dot_normalised == "localhost" or is_an_ip_address(no_dot_normalised):
        return no_dot_normalised

    if no_dot_normalised[0] == ".":
        return no_dot_normalised[1:]

    return no_dot_normalised


def normalise_same_site(same_site: str) -> Literal["strict", "lax", "none"]:
    same_site = same_site.strip()
    same_site = same_site.lower()
    allowed_values = {"strict", "lax", "none"}
    if same_site not in allowed_values:
        raise Exception('cookie same site must be one of "strict", "lax", or "none"')
    return same_site  # type: ignore


def get_url_scheme(url: str) -> str:
    url_obj = urlparse(url)
    return url_obj.scheme


class ErrorHandlers:
    def __init__(
        self,
        on_token_theft_detected: Callable[
            [BaseRequest, str, str, BaseResponse],
            Union[BaseResponse, Awaitable[BaseResponse]],
        ],
        on_try_refresh_token: Callable[
            [BaseRequest, str, BaseResponse],
            Union[BaseResponse, Awaitable[BaseResponse]],
        ],
        on_unauthorised: Callable[
            [BaseRequest, str, BaseResponse],
            Union[BaseResponse, Awaitable[BaseResponse]],
        ],
        on_invalid_claim: Callable[
            [BaseRequest, List[ClaimValidationError], BaseResponse],
            Union[BaseResponse, Awaitable[BaseResponse]],
        ],
    ):
        self.__on_token_theft_detected = on_token_theft_detected
        self.__on_try_refresh_token = on_try_refresh_token
        self.__on_unauthorised = on_unauthorised
        self.__on_invalid_claim = on_invalid_claim

    async def on_token_theft_detected(
        self,
        recipe: SessionRecipe,
        request: BaseRequest,
        session_handle: str,
        user_id: str,
        response: BaseResponse,
    ) -> BaseResponse:
        log_debug_message("Clearing tokens because of TOKEN_THEFT_DETECTED response")
        clear_session_from_all_token_transfer_methods(response, recipe)
        return await resolve(
            self.__on_token_theft_detected(request, session_handle, user_id, response)
        )

    async def on_try_refresh_token(
        self, request: BaseRequest, message: str, response: BaseResponse
    ):
        result = await resolve(self.__on_try_refresh_token(request, message, response))
        return result

    async def on_unauthorised(
        self,
        recipe: SessionRecipe,
        do_clear_cookies: bool,
        request: BaseRequest,
        message: str,
        response: BaseResponse,
    ):
        if do_clear_cookies:
            log_debug_message("Clearing tokens because of UNAUTHORISED response")
            clear_session_from_all_token_transfer_methods(response, recipe)
        return await resolve(self.__on_unauthorised(request, message, response))

    async def on_invalid_claim(
        self,
        recipe: SessionRecipe,
        request: BaseRequest,
        claim_validation_errors: List[ClaimValidationError],
        response: BaseResponse,
    ):
        _ = recipe
        result = await resolve(
            self.__on_invalid_claim(request, claim_validation_errors, response)
        )
        return result


class InputErrorHandlers(ErrorHandlers):
    def __init__(
        self,
        on_token_theft_detected: Union[
            None,
            Callable[
                [BaseRequest, str, str, BaseResponse],
                Union[BaseResponse, Awaitable[BaseResponse]],
            ],
        ] = None,
        on_unauthorised: Union[
            Callable[
                [BaseRequest, str, BaseResponse],
                Union[BaseResponse, Awaitable[BaseResponse]],
            ],
            None,
        ] = None,
        on_invalid_claim: Union[
            Callable[
                [BaseRequest, List[ClaimValidationError], BaseResponse],
                Union[BaseResponse, Awaitable[BaseResponse]],
            ],
            None,
        ] = None,
    ):
        if on_token_theft_detected is None:
            on_token_theft_detected = default_token_theft_detected_callback
        if on_unauthorised is None:
            on_unauthorised = default_unauthorised_callback
        if on_invalid_claim is None:
            on_invalid_claim = default_invalid_claim_callback
        super().__init__(
            on_token_theft_detected,
            default_try_refresh_token_callback,
            on_unauthorised,
            on_invalid_claim,
        )


async def default_unauthorised_callback(
    _: BaseRequest, __: str, response: BaseResponse
) -> BaseResponse:
    from .recipe import SessionRecipe

    return send_non_200_response_with_message(
        "unauthorised",
        SessionRecipe.get_instance().config.session_expired_status_code,
        response,
    )


async def default_try_refresh_token_callback(
    _: BaseRequest, __: str, response: BaseResponse
) -> BaseResponse:
    from .recipe import SessionRecipe

    return send_non_200_response_with_message(
        "try refresh token",
        SessionRecipe.get_instance().config.session_expired_status_code,
        response,
    )


async def default_token_theft_detected_callback(
    _: BaseRequest, session_handle: str, __: str, response: BaseResponse
) -> BaseResponse:
    from .recipe import SessionRecipe

    await SessionRecipe.get_instance().recipe_implementation.revoke_session(
        session_handle, {}
    )
    return send_non_200_response_with_message(
        "token theft detected",
        SessionRecipe.get_instance().config.session_expired_status_code,
        response,
    )


async def default_invalid_claim_callback(
    _: BaseRequest,
    claim_validation_errors: List[ClaimValidationError],
    response: BaseResponse,
) -> BaseResponse:
    from .recipe import SessionRecipe

    return send_non_200_response(
        {
            "message": "invalid claim",
            "claimValidationErrors": [err.to_json() for err in claim_validation_errors],
        },
        SessionRecipe.get_instance().config.invalid_claim_status_code,
        response,
    )


def get_auth_mode_from_header(request: BaseRequest) -> Optional[str]:
    auth_mode = request.get_header(AUTH_MODE_HEADER_KEY)
    if auth_mode is None:
        return None
    return auth_mode.lower()


def get_token_transfer_method_default(
    req: BaseRequest,
    for_create_new_session: bool,
    user_context: Dict[str, Any],
):
    _ = user_context

    # We allow fallback (checking headers then cookies) by default when validating
    if not for_create_new_session:
        return "any"

    auth_mode = get_auth_mode_from_header(req)
    if auth_mode in ("header", "cookie"):
        return auth_mode

    return "any"


class InputOverrideConfig:
    def __init__(
        self,
        functions: Union[Callable[[RecipeInterface], RecipeInterface], None] = None,
        apis: Union[Callable[[APIInterface], APIInterface], None] = None,
        openid_feature: Union[OpenIdInputOverrideConfig, None] = None,
    ):
        self.functions = functions
        self.apis = apis
        self.openid_feature = openid_feature


class OverrideConfig:
    def __init__(
        self,
        functions: Union[Callable[[RecipeInterface], RecipeInterface], None] = None,
        apis: Union[Callable[[APIInterface], APIInterface], None] = None,
    ):
        self.functions = functions
        self.apis = apis


class JWTConfig:
    def __init__(
        self,
        enable: bool,
        property_name_in_access_token_payload: Union[str, None] = None,
        issuer: Union[str, None] = None,
    ):
        if property_name_in_access_token_payload is None:
            property_name_in_access_token_payload = "jwt"
        if (
            property_name_in_access_token_payload
            == ACCESS_TOKEN_PAYLOAD_JWT_PROPERTY_NAME_KEY
        ):
            raise Exception(JWT_RESERVED_KEY_USE_ERROR_MESSAGE)
        self.enable = enable
        self.property_name_in_access_token_payload = (
            property_name_in_access_token_payload
        )
        self.issuer = issuer


TokenType = Literal["access", "refresh"]
TokenTransferMethod = Literal["cookie", "header"]


class SessionConfig:
    def __init__(
        self,
        refresh_token_path: NormalisedURLPath,
        cookie_domain: Union[None, str],
        cookie_same_site: Literal["lax", "strict", "none"],
        cookie_secure: bool,
        session_expired_status_code: int,
        error_handlers: ErrorHandlers,
        anti_csrf: str,
        get_token_transfer_method: Callable[
            [BaseRequest, bool, Dict[str, Any]],
            Union[TokenTransferMethod, Literal["any"]],
        ],
        override: OverrideConfig,
        framework: str,
        mode: str,
        jwt: JWTConfig,
        invalid_claim_status_code: int,
    ):
        self.session_expired_status_code = session_expired_status_code
        self.invalid_claim_status_code = invalid_claim_status_code

        self.refresh_token_path = refresh_token_path
        self.cookie_domain = cookie_domain
        self.cookie_same_site = cookie_same_site
        self.cookie_secure = cookie_secure
        self.error_handlers = error_handlers
        self.anti_csrf = anti_csrf
        self.get_token_transfer_method = get_token_transfer_method
        self.override = override
        self.framework = framework
        self.mode = mode
        self.jwt = jwt


def validate_and_normalise_user_input(
    app_info: AppInfo,
    cookie_domain: Union[str, None] = None,
    cookie_secure: Union[bool, None] = None,
    cookie_same_site: Union[Literal["lax", "none", "strict"], None] = None,
    session_expired_status_code: Union[int, None] = None,
    anti_csrf: Union[Literal["VIA_TOKEN", "VIA_CUSTOM_HEADER", "NONE"], None] = None,
    get_token_transfer_method: Union[
        Callable[
            [BaseRequest, bool, Dict[str, Any]],
            Union[TokenTransferMethod, Literal["any"]],
        ],
        None,
    ] = None,
    error_handlers: Union[ErrorHandlers, None] = None,
    override: Union[InputOverrideConfig, None] = None,
    jwt: Union[JWTConfig, None] = None,
    invalid_claim_status_code: Union[int, None] = None,
):
    if anti_csrf not in {"VIA_TOKEN", "VIA_CUSTOM_HEADER", "NONE", None}:
        raise ValueError(
            "anti_csrf must be one of VIA_TOKEN, VIA_CUSTOM_HEADER, NONE or None"
        )

    if error_handlers is not None and not isinstance(error_handlers, ErrorHandlers):  # type: ignore
        raise ValueError("error_handlers must be an instance of ErrorHandlers or None")

    if override is not None and not isinstance(override, InputOverrideConfig):  # type: ignore
        raise ValueError("override must be an instance of InputOverrideConfig or None")

    if jwt is not None and not isinstance(jwt, JWTConfig):  # type: ignore
        raise ValueError("jwt must be an instance of JWTConfig or None")

    cookie_domain = (
        normalise_session_scope(cookie_domain) if cookie_domain is not None else None
    )
    top_level_api_domain = app_info.top_level_api_domain
    top_level_website_domain = app_info.top_level_website_domain

    api_domain_scheme = get_url_scheme(app_info.api_domain.get_as_string_dangerous())
    website_domain_scheme = get_url_scheme(
        app_info.website_domain.get_as_string_dangerous()
    )
    if cookie_same_site is not None:
        cookie_same_site = normalise_same_site(cookie_same_site)
    elif (top_level_api_domain != top_level_website_domain) or (
        api_domain_scheme != website_domain_scheme
    ):
        cookie_same_site = "none"
    else:
        cookie_same_site = "lax"

    cookie_secure = (
        cookie_secure
        if cookie_secure is not None
        else app_info.api_domain.get_as_string_dangerous().startswith("https")
    )

    session_expired_status_code = (
        session_expired_status_code if session_expired_status_code is not None else 401
    )

    invalid_claim_status_code = (
        invalid_claim_status_code if invalid_claim_status_code is not None else 403
    )

    if session_expired_status_code == invalid_claim_status_code:
        raise Exception(
            "session_expired_status_code and invalid_claim_status_code cannot be the same "
            f"({invalid_claim_status_code})"
        )

    if anti_csrf is None:
        anti_csrf = "VIA_CUSTOM_HEADER" if cookie_same_site == "none" else "NONE"

    if get_token_transfer_method is None:
        get_token_transfer_method = get_token_transfer_method_default

    if error_handlers is None:
        error_handlers = InputErrorHandlers()

    if override is None:
        override = InputOverrideConfig()

    if jwt is None:
        jwt = JWTConfig(False)

    return SessionConfig(
        app_info.api_base_path.append(NormalisedURLPath(SESSION_REFRESH)),
        cookie_domain,
        cookie_same_site,
        cookie_secure,
        session_expired_status_code,
        error_handlers,
        anti_csrf,
        get_token_transfer_method,
        OverrideConfig(override.functions, override.apis),
        app_info.framework,
        app_info.mode,
        jwt,
        invalid_claim_status_code,
    )


async def get_required_claim_validators(
    session: SessionContainer,
    override_global_claim_validators: Optional[
        Callable[
            [List[SessionClaimValidator], SessionContainer, Dict[str, Any]],
            MaybeAwaitable[List[SessionClaimValidator]],
        ]
    ],
    user_context: Dict[str, Any],
) -> List[SessionClaimValidator]:
    from .recipe import SessionRecipe

    claim_validators_added_by_other_recipes = (
        SessionRecipe.get_instance().get_claim_validators_added_by_other_recipes()
    )
    global_claim_validators = await resolve(
        SessionRecipe.get_instance().recipe_implementation.get_global_claim_validators(
            session.get_user_id(),
            claim_validators_added_by_other_recipes,
            user_context,
        )
    )

    if override_global_claim_validators is not None:
        return await resolve(
            override_global_claim_validators(
                global_claim_validators, session, user_context
            )
        )

    return global_claim_validators


async def validate_claims_in_payload(
    claim_validators: List[SessionClaimValidator],
    new_access_token_payload: Dict[str, Any],
    user_context: Dict[str, Any],
):
    validation_errors: List[ClaimValidationError] = []
    for validator in claim_validators:
        claim_validation_res = await validator.validate(
            new_access_token_payload, user_context
        )
        log_debug_message(
            "validate_claims_in_payload %s validate res %s",
            validator.id,
            json.dumps(claim_validation_res.__dict__),
        )
        if not claim_validation_res.is_valid:
            validation_errors.append(
                ClaimValidationError(validator.id, claim_validation_res.reason)
            )

    return validation_errors



import os
from typing import Any, Dict, List

import pytest
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.recipe import (
    emailpassword,
    emailverification,
    jwt,
    openid,
    passwordless,
    session,
    thirdparty,
    thirdpartyemailpassword,
    thirdpartypasswordless,
    usermetadata,
)
from supertokens_python.recipe.emailverification.interfaces import (
    GetEmailForUserIdOkResult,
)
from supertokens_python.recipe.passwordless.utils import ContactEmailOrPhoneConfig
from supertokens_python.recipe.thirdparty.provider import Provider


@pytest.mark.asyncio
async def test_init_validation_emailpassword():
    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info="AppInfo",  # type: ignore
            framework="fastapi",
            recipe_list=[
                emailpassword.init(),
            ],
        )
    assert "app_info must be an instance of InputAppInfo" == str(ex.value)

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                emailpassword.init(sign_up_feature="sign up"),  # type: ignore
            ],
        )
    assert "sign_up_feature must be of type InputSignUpFeature or None" == str(ex.value)

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                emailpassword.init(
                    reset_password_using_token_feature="reset password"  # type: ignore
                ),
            ],
        )
    assert (
        "reset_password_using_token_feature must be of type InputResetPasswordUsingTokenFeature or None"
        == str(ex.value)
    )

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                emailverification.init("email verify"),  # type: ignore
                emailpassword.init(),
            ],
        )
    assert (
        "Email Verification recipe mode must be one of 'REQUIRED' or 'OPTIONAL'"
        == str(ex.value)
    )

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                emailpassword.init(override="override"),  # type: ignore
            ],
        )
    assert "override must be of type InputOverrideConfig or None" == str(ex.value)


async def get_email_for_user_id(_: str, __: Dict[str, Any]):
    return GetEmailForUserIdOkResult("foo@example.com")


@pytest.mark.asyncio
async def test_init_validation_emailverification():
    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[emailverification.init("config")],  # type: ignore
        )
    assert (
        "Email Verification recipe mode must be one of 'REQUIRED' or 'OPTIONAL'"
        == str(ex.value)
    )

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                emailverification.init(
                    mode="OPTIONAL",
                    get_email_for_user_id=get_email_for_user_id,
                    override="override",  # type: ignore
                )
            ],
        )
    assert "override must be of type OverrideConfig or None" == str(ex.value)


@pytest.mark.asyncio
async def test_init_validation_jwt():
    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[jwt.init(jwt_validity_seconds="100")],  # type: ignore
        )
    assert "jwt_validity_seconds must be an integer or None" == str(ex.value)

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[jwt.init(override="override")],  # type: ignore
        )
    assert "override must be an instance of OverrideConfig or None" == str(ex.value)


@pytest.mark.asyncio
async def test_init_validation_openid():
    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[openid.init(override="override")],  # type: ignore
        )
    assert "override must be an instance of InputOverrideConfig or None" == str(
        ex.value
    )


async def send_text_message(
    _: passwordless.CreateAndSendCustomTextMessageParameters, __: Dict[str, Any]
):
    pass


@pytest.mark.asyncio
async def test_init_validation_passwordless():
    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info="AppInfo",  # type: ignore
            framework="fastapi",
            recipe_list=[
                passwordless.init(
                    flow_type="USER_INPUT_CODE",
                    contact_config=passwordless.ContactPhoneOnlyConfig(
                        create_and_send_custom_text_message=send_text_message
                    ),
                )
            ],
        )
    assert "app_info must be an instance of InputAppInfo" == str(ex.value)

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                passwordless.init(
                    flow_type="SOME_OTHER_CODE",  # type: ignore
                    contact_config=passwordless.ContactPhoneOnlyConfig(
                        create_and_send_custom_text_message=send_text_message
                    ),
                )
            ],
        )
    assert (
        "flow_type must be one of USER_INPUT_CODE, MAGIC_LINK, USER_INPUT_CODE_AND_MAGIC_LINK"
        == str(ex.value)
    )

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                passwordless.init(
                    flow_type="USER_INPUT_CODE",
                    contact_config="contact config",  # type: ignore
                )
            ],
        )
    assert "contact_config must be of type ContactConfig" == str(ex.value)

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                passwordless.init(
                    flow_type="USER_INPUT_CODE",
                    contact_config=passwordless.ContactPhoneOnlyConfig(
                        create_and_send_custom_text_message=send_text_message
                    ),
                    override="override",  # type: ignore
                )
            ],
        )
    assert "override must be of type OverrideConfig" == str(ex.value)


providers_list: List[Provider] = [
    thirdparty.Google(
        client_id=os.environ.get("GOOGLE_CLIENT_ID"),  # type: ignore
        client_secret=os.environ.get("GOOGLE_CLIENT_SECRET"),  # type: ignore
    ),
    thirdparty.Facebook(
        client_id=os.environ.get("FACEBOOK_CLIENT_ID"),  # type: ignore
        client_secret=os.environ.get("FACEBOOK_CLIENT_SECRET"),  # type: ignore
    ),
    thirdparty.Github(
        client_id=os.environ.get("GITHUB_CLIENT_ID"),  # type: ignore
        client_secret=os.environ.get("GITHUB_CLIENT_SECRET"),  # type: ignore
    ),
]


@pytest.mark.asyncio
async def test_init_validation_session():
    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[session.init(anti_csrf="ABCDE")],  # type: ignore
        )
    assert "anti_csrf must be one of VIA_TOKEN, VIA_CUSTOM_HEADER, NONE or None" == str(
        ex.value
    )

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[session.init(error_handlers="error handlers")],  # type: ignore
        )
    assert "error_handlers must be an instance of ErrorHandlers or None" == str(
        ex.value
    )

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[session.init(override="override")],  # type: ignore
        )
    assert "override must be an instance of InputOverrideConfig or None" == str(
        ex.value
    )

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[session.init(jwt="jwt")],  # type: ignore
        )
    assert "jwt must be an instance of JWTConfig or None" == str(ex.value)


@pytest.mark.asyncio
async def test_init_validation_thirdparty():
    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                thirdparty.init(sign_in_and_up_feature="sign in up")  # type: ignore
            ],
        )
    assert "sign_in_and_up_feature must be an instance of SignInAndUpFeature" == str(
        ex.value
    )

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                thirdparty.init(
                    sign_in_and_up_feature=thirdparty.SignInAndUpFeature(
                        providers_list
                    ),
                    override="override",  # type: ignore
                )
            ],
        )
    assert "override must be an instance of InputOverrideConfig or None" == str(
        ex.value
    )


@pytest.mark.asyncio
async def test_init_validation_thirdpartyemailpassword():
    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                thirdpartyemailpassword.init(sign_up_feature="sign up")  # type: ignore
            ],
        )
    assert "sign_up_feature must be of type InputSignUpFeature or None" == str(ex.value)

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                thirdpartyemailpassword.init(
                    reset_password_using_token_feature="reset password"  # type: ignore
                )
            ],
        )
    assert (
        "reset_password_using_token_feature must be of type InputResetPasswordUsingTokenFeature or None"
        == str(ex.value)
    )

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                emailverification.init("email verification"),  # type: ignore
                thirdpartyemailpassword.init(),
            ],
        )
    assert (
        "Email Verification recipe mode must be one of 'REQUIRED' or 'OPTIONAL'"
        == str(ex.value)
    )

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                thirdpartyemailpassword.init(override="override")  # type: ignore
            ],
        )
    assert "override must be of type InputOverrideConfig or None" == str(ex.value)

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                thirdpartyemailpassword.init(providers="providers")  # type: ignore
            ],
        )
    assert "providers must be of type List[Provider] or None" == str(ex.value)

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                thirdpartyemailpassword.init(providers=["providers"])  # type: ignore
            ],
        )
    assert "providers must be of type List[Provider] or None" == str(ex.value)


async def save_code_text(
    _param: passwordless.CreateAndSendCustomTextMessageParameters, _: Dict[str, Any]
):
    pass


async def save_code_email(
    _param: passwordless.CreateAndSendCustomEmailParameters, _: Dict[str, Any]
):
    pass


@pytest.mark.asyncio
async def test_init_validation_thirdpartypasswordless():
    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                thirdpartypasswordless.init(
                    contact_config="contact config",  # type: ignore
                    flow_type="USER_INPUT_CODE_AND_MAGIC_LINK",
                )
            ],
        )
    assert "contact_config must be an instance of ContactConfig" == str(ex.value)

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                thirdpartypasswordless.init(
                    contact_config=ContactEmailOrPhoneConfig(
                        create_and_send_custom_text_message=save_code_text,
                        create_and_send_custom_email=save_code_email,
                    ),
                    flow_type="CUSTOM",  # type: ignore
                )
            ],
        )
    assert (
        "flow_type must be one of USER_INPUT_CODE, MAGIC_LINK, USER_INPUT_CODE_AND_MAGIC_LINK"
        == str(ex.value)
    )

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                emailverification.init(
                    "email verify",  # type: ignore
                ),
                thirdpartypasswordless.init(
                    contact_config=ContactEmailOrPhoneConfig(
                        create_and_send_custom_text_message=save_code_text,
                        create_and_send_custom_email=save_code_email,
                    ),
                    flow_type="USER_INPUT_CODE_AND_MAGIC_LINK",
                ),
            ],
        )
    assert (
        "Email Verification recipe mode must be one of 'REQUIRED' or 'OPTIONAL'"
        == str(ex.value)
    )

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                thirdpartypasswordless.init(
                    contact_config=ContactEmailOrPhoneConfig(
                        create_and_send_custom_text_message=save_code_text,
                        create_and_send_custom_email=save_code_email,
                    ),
                    flow_type="USER_INPUT_CODE_AND_MAGIC_LINK",
                    override="override",  # type: ignore
                )
            ],
        )
    assert "override must be an instance of InputOverrideConfig or None" == str(
        ex.value
    )

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                thirdpartypasswordless.init(
                    contact_config=ContactEmailOrPhoneConfig(
                        create_and_send_custom_text_message=save_code_text,
                        create_and_send_custom_email=save_code_email,
                    ),
                    flow_type="USER_INPUT_CODE_AND_MAGIC_LINK",
                    providers="providers",  # type: ignore
                )
            ],
        )
    assert "providers must be of type List[Provider] or None" == str(ex.value)

    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[
                thirdpartypasswordless.init(
                    contact_config=ContactEmailOrPhoneConfig(
                        create_and_send_custom_text_message=save_code_text,
                        create_and_send_custom_email=save_code_email,
                    ),
                    flow_type="USER_INPUT_CODE_AND_MAGIC_LINK",
                    providers=["providers"],  # type: ignore
                )
            ],
        )
    assert "providers must be of type List[Provider] or None" == str(ex.value)


@pytest.mark.asyncio
async def test_init_validation_usermetadata():
    with pytest.raises(ValueError) as ex:
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="fastapi",
            recipe_list=[usermetadata.init(override="override")],  # type: ignore
        )
    assert "override must be an instance of InputOverrideConfig or None" == str(
        ex.value
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Union, Optional

from supertokens_python.recipe.userroles.interfaces import APIInterface, RecipeInterface
from supertokens_python.supertokens import AppInfo

if TYPE_CHECKING:
    from supertokens_python.recipe.userroles.recipe import UserRolesRecipe


class InputOverrideConfig:
    def __init__(
        self,
        functions: Union[Callable[[RecipeInterface], RecipeInterface], None] = None,
        apis: Union[Callable[[APIInterface], APIInterface], None] = None,
    ):
        self.functions = functions
        self.apis = apis


class UserRolesConfig:
    def __init__(
        self,
        skip_adding_roles_to_access_token: bool,
        skip_adding_permissions_to_access_token: bool,
        override: InputOverrideConfig,
    ) -> None:
        self.skip_adding_roles_to_access_token = skip_adding_roles_to_access_token
        self.skip_adding_permissions_to_access_token = (
            skip_adding_permissions_to_access_token
        )
        self.override = override


def validate_and_normalise_user_input(
    _recipe: UserRolesRecipe,
    _app_info: AppInfo,
    skip_adding_roles_to_access_token: Optional[bool] = None,
    skip_adding_permissions_to_access_token: Optional[bool] = None,
    override: Union[InputOverrideConfig, None] = None,
) -> UserRolesConfig:
    if override is not None and not isinstance(override, InputOverrideConfig):  # type: ignore
        raise ValueError("override must be an instance of InputOverrideConfig or None")

    if override is None:
        override = InputOverrideConfig()

    if skip_adding_roles_to_access_token is None:
        skip_adding_roles_to_access_token = False
    if skip_adding_permissions_to_access_token is None:
        skip_adding_permissions_to_access_token = False

    return UserRolesConfig(
        skip_adding_roles_to_access_token=skip_adding_roles_to_access_token,
        skip_adding_permissions_to_access_token=skip_adding_permissions_to_access_token,
        override=override,
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Union

from supertokens_python.framework import BaseResponse

if TYPE_CHECKING:
    from fastapi import FastAPI, Request


def get_middleware():
    from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint

    class Middleware(BaseHTTPMiddleware):
        def __init__(self, app: FastAPI):
            super().__init__(app)

        async def dispatch(self, request: Request, call_next: RequestResponseEndpoint):
            from supertokens_python import Supertokens
            from supertokens_python.exceptions import SuperTokensError
            from supertokens_python.framework.fastapi.fastapi_request import (
                FastApiRequest,
            )
            from supertokens_python.framework.fastapi.fastapi_response import (
                FastApiResponse,
            )
            from supertokens_python.recipe.session import SessionContainer
            from supertokens_python.supertokens import manage_session_post_response

            st = Supertokens.get_instance()
            from fastapi.responses import Response

            try:
                custom_request = FastApiRequest(request)
                response = FastApiResponse(Response())
                result: Union[BaseResponse, None] = await st.middleware(
                    custom_request, response
                )
                if result is None:
                    response = await call_next(request)
                    result = FastApiResponse(response)

                if hasattr(request.state, "supertokens") and isinstance(
                    request.state.supertokens, SessionContainer
                ):
                    manage_session_post_response(request.state.supertokens, result)
                if isinstance(result, FastApiResponse):
                    return result.response
            except SuperTokensError as e:
                response = FastApiResponse(Response())
                result: Union[BaseResponse, None] = await st.handle_supertokens_error(
                    FastApiRequest(request), e, response
                )
                if isinstance(result, FastApiResponse):
                    return result.response

            raise Exception("Should never come here")

    return Middleware



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Any, Dict, List, Union

from supertokens_python.recipe.emailpassword.interfaces import (
    APIInterface,
    APIOptions,
    SignInPostOkResult,
    SignInPostWrongCredentialsError,
    SignUpPostEmailAlreadyExistsError,
    SignUpPostOkResult,
)
from supertokens_python.recipe.emailpassword.types import FormField, User
from supertokens_python.recipe.thirdpartyemailpassword.interfaces import (
    APIInterface as ThirdPartyEmailPasswordAPIInterface,
)
from supertokens_python.recipe.thirdpartyemailpassword.interfaces import (
    EmailPasswordSignInPostOkResult,
    EmailPasswordSignUpPostOkResult,
)
from supertokens_python.types import GeneralErrorResponse


def get_interface_impl(
    api_implementation: ThirdPartyEmailPasswordAPIInterface,
) -> APIInterface:
    implementation = APIInterface()

    implementation.disable_email_exists_get = (
        api_implementation.disable_email_exists_get
    )
    implementation.disable_generate_password_reset_token_post = (
        api_implementation.disable_generate_password_reset_token_post
    )
    implementation.disable_password_reset_post = (
        api_implementation.disable_password_reset_post
    )
    implementation.disable_sign_in_post = (
        api_implementation.disable_emailpassword_sign_in_post
    )
    implementation.disable_sign_up_post = (
        api_implementation.disable_emailpassword_sign_up_post
    )

    implementation.email_exists_get = api_implementation.emailpassword_email_exists_get
    implementation.generate_password_reset_token_post = (
        api_implementation.generate_password_reset_token_post
    )
    implementation.password_reset_post = api_implementation.password_reset_post
    if not implementation.disable_sign_in_post:

        async def sign_in_post(
            form_fields: List[FormField],
            api_options: APIOptions,
            user_context: Dict[str, Any],
        ) -> Union[
            SignInPostOkResult, SignInPostWrongCredentialsError, GeneralErrorResponse
        ]:
            result = await api_implementation.emailpassword_sign_in_post(
                form_fields, api_options, user_context
            )
            if isinstance(result, EmailPasswordSignInPostOkResult):
                return SignInPostOkResult(
                    User(
                        result.user.user_id, result.user.email, result.user.time_joined
                    ),
                    result.session,
                )
            return result

        implementation.sign_in_post = sign_in_post

    if not implementation.disable_sign_up_post:

        async def sign_up_post(
            form_fields: List[FormField],
            api_options: APIOptions,
            user_context: Dict[str, Any],
        ) -> Union[
            SignUpPostOkResult, SignUpPostEmailAlreadyExistsError, GeneralErrorResponse
        ]:
            result = await api_implementation.emailpassword_sign_up_post(
                form_fields, api_options, user_context
            )
            if isinstance(result, EmailPasswordSignUpPostOkResult):
                return SignUpPostOkResult(
                    User(
                        result.user.user_id, result.user.email, result.user.time_joined
                    ),
                    result.session,
                )
            return result

        implementation.sign_up_post = sign_up_post

    return implementation



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import asyncio
import json
from typing import Any, Dict, Union

from fastapi import FastAPI
from fastapi.requests import Request
from fastapi.testclient import TestClient
from pytest import fixture, mark
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.recipe import emailpassword, session
from supertokens_python.recipe.emailpassword.types import User
from supertokens_python.recipe.session import SessionContainer
from supertokens_python.recipe.session.asyncio import (
    create_new_session,
    get_session,
    refresh_session,
)
from tests.utils import clean_st, reset, setup_st, sign_up_request, start_st


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@fixture(scope="function")
async def driver_config_client():
    app = FastAPI()
    app.add_middleware(get_middleware())

    @app.get("/login")
    async def login(request: Request):  # type: ignore
        user_id = "userId"
        await create_new_session(request, user_id, {}, {})
        return {"userId": user_id}

    @app.post("/refresh")
    async def custom_refresh(request: Request):  # type: ignore
        await refresh_session(request)
        return {}  # type: ignore

    @app.get("/info")
    async def info_get(request: Request):  # type: ignore
        await get_session(request, True)
        return {}  # type: ignore

    @app.get("/custom/info")
    def custom_info(_):  # type: ignore
        return {}  # type: ignore

    @app.options("/custom/handle")
    def custom_handle_options(_):  # type: ignore
        return {"method": "option"}

    @app.get("/handle")
    async def handle_get(request: Request):  # type: ignore
        session: Union[None, SessionContainer] = await get_session(request, True)
        if session is None:
            raise Exception("Should never come here")
        return {"s": session.get_handle()}

    @app.post("/logout")
    async def custom_logout(request: Request):  # type: ignore
        session: Union[None, SessionContainer] = await get_session(request, True)
        if session is None:
            raise Exception("Should never come here")
        await session.revoke_session()
        return {}  # type: ignore

    return TestClient(app)


@mark.asyncio
async def test_email_validation_checks_in_generate_token_API(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[emailpassword.init()],
    )
    start_st()

    for invalid_email in ["random", 5]:
        res = driver_config_client.post(
            url="/auth/user/password/reset/token",
            json={"formFields": [{"id": "email", "value": invalid_email}]},
        )

        assert res.status_code == 200
        dict_res = json.loads(res.text)
        assert dict_res["status"] == "FIELD_ERROR"


@mark.asyncio
async def test_that_generated_password_link_is_correct(
    driver_config_client: TestClient,
):
    reset_url = None
    token_info: Union[None, str] = None
    rid_info: Union[None, str] = None

    async def custom_f(_: User, password_reset_url_with_token: str, __: Dict[str, Any]):
        nonlocal reset_url, token_info, rid_info
        reset_url = password_reset_url_with_token.split("?")[0]
        token_info = password_reset_url_with_token.split("?")[1].split("&")[0]
        rid_info = password_reset_url_with_token.split("?")[1].split("&")[1]

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
            emailpassword.init(
                reset_password_using_token_feature=emailpassword.InputResetPasswordUsingTokenFeature(
                    create_and_send_custom_email=custom_f
                )
            ),
        ],
    )
    start_st()

    response_1 = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"
    await asyncio.sleep(1)
    response_1 = driver_config_client.post(
        url="/auth/user/password/reset/token",
        json={"formFields": [{"id": "email", "value": "test@gmail.com"}]},
    )
    await asyncio.sleep(1)

    assert response_1.status_code == 200
    assert reset_url == "http://supertokens.io/auth/reset-password"
    assert token_info is not None and "token=" in token_info  # type: ignore pylint: disable=unsupported-membership-test
    assert rid_info is not None and "rid=emailpassword" in rid_info  # type: ignore pylint: disable=unsupported-membership-test


@mark.asyncio
async def test_password_validation(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[emailpassword.init()],
    )
    start_st()

    response_1 = driver_config_client.post(
        url="/auth/user/password/reset",
        json={
            "formFields": [{"id": "password", "value": "invalid"}],
            "token": "random",
        },
    )

    assert response_1.status_code == 200
    # dict_response = json.loads(response_1.text)
    # assert dict_response["status"] == "FIELD_ERROR"

    response_2 = driver_config_client.post(
        url="/auth/user/password/reset",
        json={
            "formFields": [{"id": "password", "value": "validpass123"}],
            "token": "randomToken",
        },
    )

    assert response_2.status_code == 200
    dict_response = json.loads(response_2.text)
    assert dict_response["status"] != "FIELD_ERROR"


@mark.asyncio
async def test_token_missing_from_input(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[emailpassword.init()],
    )
    start_st()

    response_1 = driver_config_client.post(
        url="/auth/user/password/reset",
        json={"formFields": [{"id": "password", "value": "validpass123"}]},
    )

    assert response_1.status_code == 400
    dict_response = json.loads(response_1.text)
    assert dict_response["message"] == "Please provide the password reset token"


@mark.asyncio
async def test_valid_token_input_and_passoword_has_changed(
    driver_config_client: TestClient,
):
    token_info = None

    async def custom_f(_: User, password_reset_url_with_token: str, __: Dict[str, Any]):
        nonlocal token_info
        token_info = (
            password_reset_url_with_token.split("?")[1].split("&")[0].split("=")[1]
        )

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(
                reset_password_using_token_feature=emailpassword.InputResetPasswordUsingTokenFeature(
                    create_and_send_custom_email=custom_f
                )
            ),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validpass123"
    )
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    user_info = dict_response["user"]
    assert dict_response["status"] == "OK"
    response_1 = driver_config_client.post(
        url="/auth/user/password/reset/token",
        json={"formFields": [{"id": "email", "value": "random@gmail.com"}]},
    )
    await asyncio.sleep(1)

    assert response_1.status_code == 200

    response_2 = driver_config_client.post(
        url="/auth/user/password/reset",
        json={
            "formFields": [{"id": "password", "value": "validpass12345"}],
            "token": token_info,
        },
    )
    await asyncio.sleep(1)
    assert response_2.status_code == 200

    response_3 = driver_config_client.post(
        url="/auth/signin",
        json={
            "formFields": [
                {"id": "password", "value": "validpass123"},
                {"id": "email", "value": "random@gmail.com"},
            ]
        },
    )

    assert response_3.status_code == 200

    dict_response = json.loads(response_3.text)
    assert dict_response["status"] == "WRONG_CREDENTIALS_ERROR"

    response_4 = driver_config_client.post(
        url="/auth/signin",
        json={
            "formFields": [
                {
                    "id": "password",
                    "value": "validpass12345",
                },
                {
                    "id": "email",
                    "value": "random@gmail.com",
                },
            ]
        },
    )

    assert response_4.status_code == 200

    dict_response = json.loads(response_4.text)
    assert dict_response["status"] == "OK"
    assert dict_response["user"]["id"] == user_info["id"]
    assert dict_response["user"]["email"] == user_info["email"]



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Callable, List, Union

from supertokens_python.ingredients.emaildelivery.types import (
    EmailDeliveryConfig,
    EmailDeliveryConfigWithService,
)
from supertokens_python.recipe.emailpassword.interfaces import (
    RecipeInterface as EPRecipeInterface,
)
from supertokens_python.recipe.thirdparty.provider import Provider

from ..emailpassword.utils import (
    InputResetPasswordUsingTokenFeature,
    InputSignUpFeature,
)
from .emaildelivery.services.backward_compatibility import BackwardCompatibilityService
from .interfaces import APIInterface, RecipeInterface
from .types import EmailTemplateVars

if TYPE_CHECKING:
    from .recipe import ThirdPartyEmailPasswordRecipe


class InputOverrideConfig:
    def __init__(
        self,
        functions: Union[Callable[[RecipeInterface], RecipeInterface], None] = None,
        apis: Union[Callable[[APIInterface], APIInterface], None] = None,
    ):
        self.functions = functions
        self.apis = apis


class OverrideConfig:
    def __init__(
        self,
        functions: Union[Callable[[RecipeInterface], RecipeInterface], None] = None,
        apis: Union[Callable[[APIInterface], APIInterface], None] = None,
    ):
        self.functions = functions
        self.apis = apis


class ThirdPartyEmailPasswordConfig:
    def __init__(
        self,
        providers: List[Provider],
        sign_up_feature: Union[InputSignUpFeature, None],
        reset_password_using_token_feature: Union[
            InputResetPasswordUsingTokenFeature, None
        ],
        get_email_delivery_config: Callable[
            [EPRecipeInterface],
            EmailDeliveryConfigWithService[EmailTemplateVars],
        ],
        override: OverrideConfig,
    ):
        self.sign_up_feature = sign_up_feature
        self.providers = providers
        self.reset_password_using_token_feature = reset_password_using_token_feature
        self.get_email_delivery_config = get_email_delivery_config
        self.override = override


def validate_and_normalise_user_input(
    recipe: ThirdPartyEmailPasswordRecipe,
    sign_up_feature: Union[InputSignUpFeature, None] = None,
    reset_password_using_token_feature: Union[
        InputResetPasswordUsingTokenFeature, None
    ] = None,
    override: Union[InputOverrideConfig, None] = None,
    providers: Union[List[Provider], None] = None,
    email_delivery: Union[EmailDeliveryConfig[EmailTemplateVars], None] = None,
) -> ThirdPartyEmailPasswordConfig:
    if sign_up_feature is not None and not isinstance(sign_up_feature, InputSignUpFeature):  # type: ignore
        raise ValueError("sign_up_feature must be of type InputSignUpFeature or None")

    if reset_password_using_token_feature is not None and not isinstance(reset_password_using_token_feature, InputResetPasswordUsingTokenFeature):  # type: ignore
        raise ValueError(
            "reset_password_using_token_feature must be of type InputResetPasswordUsingTokenFeature or None"
        )

    if override is not None and not isinstance(override, InputOverrideConfig):  # type: ignore
        raise ValueError("override must be of type InputOverrideConfig or None")

    if providers is not None and not isinstance(providers, List):  # type: ignore
        raise ValueError("providers must be of type List[Provider] or None")

    for provider in providers or []:
        if not isinstance(provider, Provider):  # type: ignore
            raise ValueError("providers must be of type List[Provider] or None")

    if providers is None:
        providers = []
    if override is None:
        override = InputOverrideConfig()

    def get_email_delivery_config(
        ep_recipe_interface_impl: EPRecipeInterface,
    ):
        if email_delivery and email_delivery.service:
            return EmailDeliveryConfigWithService(
                service=email_delivery.service, override=email_delivery.override
            )

        email_service = BackwardCompatibilityService(
            app_info=recipe.app_info,
            ep_recipe_interface_impl=ep_recipe_interface_impl,
            reset_password_using_token_feature=reset_password_using_token_feature,
        )
        if email_delivery is not None and email_delivery.override is not None:
            override = email_delivery.override
        else:
            override = None

        return EmailDeliveryConfigWithService(email_service, override=override)

    return ThirdPartyEmailPasswordConfig(
        providers,
        sign_up_feature,
        reset_password_using_token_feature,
        get_email_delivery_config,
        OverrideConfig(functions=override.functions, apis=override.apis),
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Union

from supertokens_python.normalised_url_path import NormalisedURLPath
from supertokens_python.querier import Querier

if TYPE_CHECKING:
    from .utils import JWTConfig
    from supertokens_python.supertokens import AppInfo

from supertokens_python.recipe.jwt.interfaces import (
    CreateJwtOkResult,
    CreateJwtResultUnsupportedAlgorithm,
    GetJWKSResult,
    RecipeInterface,
)

from .interfaces import JsonWebKey


class RecipeImplementation(RecipeInterface):
    def __init__(self, querier: Querier, config: JWTConfig, app_info: AppInfo):
        super().__init__()
        self.querier = querier
        self.config = config
        self.app_info = app_info

    async def create_jwt(
        self,
        payload: Dict[str, Any],
        validity_seconds: Union[int, None],
        user_context: Dict[str, Any],
    ) -> Union[CreateJwtOkResult, CreateJwtResultUnsupportedAlgorithm]:
        if validity_seconds is None:
            validity_seconds = self.config.jwt_validity_seconds

        data = {
            "payload": payload,
            "validity": validity_seconds,
            "algorithm": "RS256",
            "jwksDomain": self.app_info.api_domain.get_as_string_dangerous(),
        }
        response = await self.querier.send_post_request(
            NormalisedURLPath("/recipe/jwt"), data
        )

        if response["status"] == "OK":
            return CreateJwtOkResult(response["jwt"])
        return CreateJwtResultUnsupportedAlgorithm()

    async def get_jwks(self, user_context: Dict[str, Any]) -> GetJWKSResult:
        response = await self.querier.send_get_request(
            NormalisedURLPath("/recipe/jwt/jwks"), {}
        )

        keys: List[JsonWebKey] = []
        for key in response["keys"]:
            keys.append(
                JsonWebKey(
                    key["kty"], key["kid"], key["n"], key["e"], key["alg"], key["use"]
                )
            )
        return GetJWKSResult(keys)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urlparse

if TYPE_CHECKING:
    pass
from .exceptions import raise_general_exception


class NormalisedURLPath:
    def __init__(self, url: str):
        self.__value = normalise_url_path_or_throw_error(url)

    def startswith(self, other: NormalisedURLPath) -> bool:
        return self.__value.startswith(other.get_as_string_dangerous())

    def append(self, other: NormalisedURLPath) -> NormalisedURLPath:
        return NormalisedURLPath(self.__value + other.get_as_string_dangerous())

    def get_as_string_dangerous(self) -> str:
        return self.__value

    def equals(self, other: NormalisedURLPath) -> bool:
        return self.__value == other.get_as_string_dangerous()

    def is_a_recipe_path(self) -> bool:
        return self.__value == "/recipe" or self.__value.startswith("/recipe/")


def normalise_url_path_or_throw_error(input_str: str) -> str:
    input_str = input_str.strip().lower()

    try:
        if (not input_str.startswith("http://")) and (
            not input_str.startswith("https://")
        ):
            raise Exception("converting to proper URL")
        url_obj = urlparse(input_str)
        input_str = url_obj.path

        if input_str.endswith("/"):
            return input_str[:-1]

        return input_str
    except Exception:
        pass

    if (
        (domain_given(input_str) or input_str.startswith("localhost"))
        and (not input_str.startswith("http://"))
        and (not input_str.startswith("https://"))
    ):
        input_str = "http://" + input_str
        return normalise_url_path_or_throw_error(input_str)

    if not input_str.startswith("/"):
        input_str = "/" + input_str

    try:
        urlparse("http://example.com" + input_str)
        return normalise_url_path_or_throw_error("http://example.com" + input_str)
    except Exception:
        raise_general_exception("Please provide a valid URL path")


def domain_given(input_str: str) -> bool:
    if ("." not in input_str) or (input_str.startswith("/")):
        return False

    try:
        url = urlparse(input_str)
        if url.hostname is None:
            raise Exception("Should never come here")
        return url.hostname.find(".") != -1
    except Exception:
        pass

    try:
        url = urlparse("http://" + input_str)
        if url.hostname is None:
            raise Exception("Should never come here")
        return url.hostname.find(".") != -1
    except Exception:
        pass

    return False



from unittest.mock import MagicMock

from pytest import mark

from supertokens_python import init
from supertokens_python.framework.request import BaseRequest
from supertokens_python.recipe import session
from supertokens_python.recipe.session.asyncio import (
    create_new_session,
    get_claim_value,
)
from supertokens_python.recipe.session.interfaces import (
    SessionContainer,
    GetClaimValueOkResult,
    SessionDoesNotExistError,
)
from tests.utils import setup_function, teardown_function, start_st, st_init_common_args
from .utils import TrueClaim, get_st_init_args

_ = setup_function  # type:ignore
_ = teardown_function  # type:ignore

pytestmark = mark.asyncio


async def test_should_get_the_right_value():
    init(**get_st_init_args(TrueClaim))  # type:ignore
    start_st()

    dummy_req: BaseRequest = MagicMock()
    s = await create_new_session(dummy_req, "someId")

    res = await s.get_claim_value(TrueClaim)
    assert res is True


async def test_should_get_the_right_value_using_session_handle():
    init(**get_st_init_args(TrueClaim))  # type:ignore
    start_st()

    dummy_req: BaseRequest = MagicMock()
    s: SessionContainer = await create_new_session(dummy_req, "someId")

    res = await get_claim_value(s.get_handle(), TrueClaim)
    assert isinstance(res, GetClaimValueOkResult)
    assert res.value is True


async def test_should_work_for_non_existing_handle():
    new_st_init = {
        **st_init_common_args,
        "recipe_list": [
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    }
    init(**new_st_init)  # type: ignore
    start_st()

    res = await get_claim_value("non_existing_handle", TrueClaim)
    assert isinstance(res, SessionDoesNotExistError)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Any
import os

import django

os.environ.setdefault("DJANGO_SETTINGS_MODULE", __file__)
SECRET_KEY = "this_is_a_random_string"

django.setup()


REST_FRAMEWORK: Any = {
    "DEFAULT_AUTHENTICATION_CLASSES": [],
    "DEFAULT_PERMISSION_CLASSES": [],
    "UNAUTHENTICATED_USER": None,
}
DATABASES = {
    "default": {
        # Add 'postgresql_psycopg2', 'postgresql', 'mysql', 'sqlite3' or
        # 'oracle'.
        "ENGINE": "django.db.backends.sqlite3",
        # Or path to database file if using sqlite3.
        "NAME": "sqlite.db",
        "USER": "",  # Not used with sqlite3.
        "PASSWORD": "",  # Not used with sqlite3.
        # Set to empty string for localhost. Not used with sqlite3.
        "HOST": "",
        # Set to empty string for default. Not used with sqlite3.
        "PORT": "",
    }
}

MIDDLEWARE = [
    "supertokens_python.framework.django.django_middleware.middleware",
]



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from .utils import TokenTransferMethod

SESSION_REFRESH = "/session/refresh"
SIGNOUT = "/signout"
ACCESS_TOKEN_COOKIE_KEY = "sAccessToken"
REFRESH_TOKEN_COOKIE_KEY = "sRefreshToken"
FRONT_TOKEN_HEADER_SET_KEY = "front-token"
ANTI_CSRF_HEADER_KEY = "anti-csrf"
RID_HEADER_KEY = "rid"
AUTH_MODE_HEADER_KEY = "st-auth-mode"
AUTHORIZATION_HEADER_KEY = "authorization"
ACCESS_TOKEN_HEADER_KEY = "st-access-token"
REFRESH_TOKEN_HEADER_KEY = "st-refresh-token"
ACCESS_CONTROL_EXPOSE_HEADERS = "Access-Control-Expose-Headers"

available_token_transfer_methods: List[TokenTransferMethod] = ["cookie", "header"]



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Dict

from pytest import fixture, mark, skip
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.querier import Querier
from supertokens_python.recipe import passwordless, session, thirdpartypasswordless
from supertokens_python.recipe.passwordless.interfaces import DeleteUserInfoOkResult
from supertokens_python.recipe.passwordless.utils import ContactEmailOrPhoneConfig
from supertokens_python.recipe.thirdpartypasswordless.asyncio import (
    delete_email_for_passwordless_user,
    delete_phone_number_for_user,
    get_user_by_id,
    get_user_by_phone_number,
    get_users_by_email,
    update_passwordless_user,
)
from supertokens_python.utils import is_version_gte

from fastapi import FastAPI
from fastapi.testclient import TestClient
from tests.utils import clean_st, reset, setup_st, start_st


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@fixture(scope="function")
async def driver_config_client():
    app = FastAPI()
    app.add_middleware(get_middleware())

    return TestClient(app)


@mark.asyncio
async def test_tp_passworldless_delete_user_info(driver_config_client: TestClient):
    DUMMY_CODE = "DUMMY_CODE"

    async def save_code_text(
        _param: passwordless.CreateAndSendCustomTextMessageParameters, _: Dict[str, Any]
    ):
        pass

    async def save_code_email(
        _param: passwordless.CreateAndSendCustomEmailParameters, _: Dict[str, Any]
    ):
        pass

    async def get_custom_user_input_code(_input: Dict[str, Any]) -> str:
        return DUMMY_CODE

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            thirdpartypasswordless.init(
                contact_config=ContactEmailOrPhoneConfig(
                    create_and_send_custom_text_message=save_code_text,
                    create_and_send_custom_email=save_code_email,
                ),
                flow_type="USER_INPUT_CODE_AND_MAGIC_LINK",
                providers=[],
                get_custom_user_input_code=get_custom_user_input_code,
            ),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    version = await Querier.get_instance().get_api_version()
    if not is_version_gte(version, "2.11"):
        # If the version less than 2.11.0, passwordless OTP doesn't exist. So skip the test
        skip()

    create_code_json = driver_config_client.post(
        url="/auth/signinup/code",
        json={
            "phoneNumber": "+919494949494",
        },
    ).json()

    consume_code_json = driver_config_client.post(
        url="/auth/signinup/code/consume",
        json={
            "preAuthSessionId": create_code_json["preAuthSessionId"],
            "deviceId": create_code_json["deviceId"],
            "userInputCode": DUMMY_CODE,
        },
    ).json()

    user_id = consume_code_json["user"]["id"]

    # Update user email as well as phone number
    await update_passwordless_user(
        user_id, email="foo@example.com", phone_number="+919494949494"
    )

    # Check that deleting the user phone number works:
    response = await delete_phone_number_for_user(user_id)
    assert isinstance(response, DeleteUserInfoOkResult)

    user = await get_user_by_phone_number("+919494949494")
    assert user is None

    user = await get_user_by_id(user_id)
    assert user is not None and user.phone_number is None

    # Restore user email and phone number
    await update_passwordless_user(
        user_id, email="foo@example.com", phone_number="+919494949494"
    )

    # Check that deleting the user email works:
    response = await delete_email_for_passwordless_user(user_id)
    assert isinstance(response, DeleteUserInfoOkResult)

    users = await get_users_by_email("foo@example.com")
    assert users == []

    user = await get_user_by_id(user_id)
    assert user is not None and user.email is None



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Union, Optional

from . import utils
from .recipe import UserRolesRecipe
from . import recipe

PermissionClaim = recipe.PermissionClaim
UserRoleClaim = recipe.UserRoleClaim

if TYPE_CHECKING:
    from supertokens_python.supertokens import AppInfo

    from ...recipe_module import RecipeModule


def init(
    skip_adding_roles_to_access_token: Optional[bool] = None,
    skip_adding_permissions_to_access_token: Optional[bool] = None,
    override: Union[utils.InputOverrideConfig, None] = None,
) -> Callable[[AppInfo], RecipeModule]:
    return UserRolesRecipe.init(
        skip_adding_roles_to_access_token,
        skip_adding_permissions_to_access_token,
        override,
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from string import Template

from supertokens_python.ingredients.emaildelivery.types import EmailContent
from supertokens_python.recipe.emailverification.types import (
    VerificationEmailTemplateVars,
)
from supertokens_python.supertokens import Supertokens

from .email_verify_email import html_template


def get_email_verify_email_content(
    email_input: VerificationEmailTemplateVars,
) -> EmailContent:
    supertokens = Supertokens.get_instance()
    app_name = supertokens.app_info.app_name
    body = get_email_verify_email_html(
        app_name, email_input.user.email, email_input.email_verify_link
    )
    return EmailContent(
        body, "Email verification instructions", email_input.user.email, is_html=True
    )


def get_email_verify_email_html(app_name: str, email: str, verification_link: str):
    return Template(html_template).substitute(
        appname=app_name, verificationLink=verification_link, toEmail=email
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
from typing import Any, Union

from asgiref.sync import async_to_sync


def middleware(get_response: Any):
    from supertokens_python import Supertokens
    from supertokens_python.exceptions import SuperTokensError
    from supertokens_python.framework.django.django_request import DjangoRequest
    from supertokens_python.framework.django.django_response import DjangoResponse
    from supertokens_python.recipe.session import SessionContainer
    from supertokens_python.supertokens import manage_session_post_response

    from django.http import HttpRequest

    if asyncio.iscoroutinefunction(get_response):

        async def __asyncMiddleware(request: HttpRequest):
            st = Supertokens.get_instance()
            custom_request = DjangoRequest(request)
            from django.http import HttpResponse

            response = DjangoResponse(HttpResponse())
            try:
                result = await st.middleware(custom_request, response)
                if result is None:
                    result = await get_response(request)
                    result = DjangoResponse(result)
                if hasattr(request, "supertokens") and isinstance(
                    request.supertokens, SessionContainer  # type: ignore
                ):
                    manage_session_post_response(
                        request.supertokens, result  # type: ignore
                    )
                if isinstance(result, DjangoResponse):
                    return result.response
            except SuperTokensError as e:
                response = DjangoResponse(HttpResponse())
                result = await st.handle_supertokens_error(
                    DjangoRequest(request), e, response
                )
                if isinstance(result, DjangoResponse):
                    return result.response

            raise Exception("Should never come here")

        return __asyncMiddleware

    def __syncMiddleware(request: HttpRequest):
        st = Supertokens.get_instance()
        custom_request = DjangoRequest(request)
        from django.http import HttpResponse

        response = DjangoResponse(HttpResponse())
        try:
            result: Union[DjangoResponse, None] = async_to_sync(st.middleware)(
                custom_request, response
            )

            if result is None:
                result = DjangoResponse(get_response(request))

            if hasattr(request, "supertokens") and isinstance(
                request.supertokens, SessionContainer  # type: ignore
            ):
                manage_session_post_response(
                    request.supertokens, result  # type: ignore
                )
            return result.response

        except SuperTokensError as e:
            response = DjangoResponse(HttpResponse())
            result: Union[DjangoResponse, None] = async_to_sync(
                st.handle_supertokens_error
            )(DjangoRequest(request), e, response)
            if result is not None:
                return result.response
        raise Exception("Should never come here")

    return __syncMiddleware



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.recipe.passwordless.interfaces import APIInterface, APIOptions
from supertokens_python.utils import default_user_context, send_200_response


async def resend_code(api_implementation: APIInterface, api_options: APIOptions):
    if api_implementation.disable_resend_code_post:
        return None
    body = await api_options.request.json()

    if body is None:
        raise_bad_input_exception("Please provide a JSON body")

    if "preAuthSessionId" not in body:
        raise_bad_input_exception("Please provide preAuthSessionId")

    if "deviceId" not in body:
        raise_bad_input_exception("Please provide deviceId")

    pre_auth_session_id = body["preAuthSessionId"]
    device_id = body["deviceId"]
    user_context = default_user_context(api_options.request)

    result = await api_implementation.resend_code_post(
        device_id, pre_auth_session_id, api_options, user_context
    )
    return send_200_response(result.to_json(), api_options.response)
